# -*- coding: utf-8 -*-
# pylint: disable=r0904, c0415
""" unittests for dkb_robo """
import sys
import os
import unittest
import logging
from unittest.mock import patch, Mock, MagicMock
from dataclasses import dataclass, asdict
sys.path.insert(0, '.')
sys.path.insert(0, '..')

@dataclass
class DataclassObject:
    Attr1: str
    attr2: int
    attr3: dict
    attr4: list

class TestDKBRobo(unittest.TestCase):
    """ test class """

    maxDiff = None

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        from dkb_robo.utilities import validate_dates, generate_random_string, logger_setup, string2float, _convert_date_format, get_dateformat, get_valid_filename, object2dictionary, logger_setup, ulal
        self.validate_dates = validate_dates
        self.string2float = string2float
        self.generate_random_string = generate_random_string
        self.logger_setup = logger_setup
        self._convert_date_format = _convert_date_format
        self.get_dateformat = get_dateformat
        self.get_valid_filename = get_valid_filename
        self.object2dictionary = object2dictionary
        self.logger = logging.getLogger('dkb_robo')
        self.logger_setup = logger_setup
        self.ulal = ulal

    @patch('time.time')
    def test_001_validate_dates(self, mock_time):
        """ test validate dates with correct data """
        date_from = '01.01.2024'
        date_to = '10.01.2024'
        mock_time.return_value = 1726309859
        self.assertEqual(('2024-01-01', '2024-01-10'), self.validate_dates(date_from, date_to))

    @patch('time.time')
    def test_002_validate_dates(self, mock_time):
        """ test validate dates dates to be corrected """
        date_from = '12.12.2021'
        date_to = '11.12.2021'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2022-01-01', '2022-01-01'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to 2022-01-01', lcm.output)
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_to to 2022-01-01', lcm.output)

    @patch('time.time')
    def test_003_validate_dates(self, mock_time):
        """ test validate dates with correct data """
        date_from = '2024-01-01'
        date_to = '2024-01-10'
        mock_time.return_value = 1726309859
        self.assertEqual(('2024-01-01', '2024-01-10'), self.validate_dates(date_from, date_to))

    @patch('time.time')
    def test_004_validate_dates(self, mock_time):
        """ test validate dates dates to be corrected """
        date_from = '2021-12-01'
        date_to = '2021-12-11'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2022-01-01', '2022-01-01'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to 2022-01-01', lcm.output)
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_to to 2022-01-01', lcm.output)

    @patch('time.time')
    def test_005_validate_dates(self, mock_time):
        """ test validate dates dates to be corrected """
        date_from = '01.12.2024'
        date_to = '11.12.2024'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2024-09-14', '2024-09-14'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to 2024-09-14', lcm.output)
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_to to 2024-09-14', lcm.output)

    @patch('time.time')
    def test_006_validate_dates(self, mock_time):
        """ test validate dates dates to be corrected """
        date_from = '2024-12-01'
        date_to = '2024-12-11'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2024-09-14', '2024-09-14'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to 2024-09-14', lcm.output)
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_to to 2024-09-14', lcm.output)

    @patch('time.time')
    def test_007_validate_dates(self, mock_time):
        """ test validate dates with correct data """
        date_from = '15.01.2024'
        date_to = '10.01.2024'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2024-01-10', '2024-01-10'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to date_to', lcm.output)

    @patch('time.time')
    def test_008_validate_dates(self, mock_time):
        """ test validate dates with correct data """
        date_from = '2024-01-15'
        date_to = '2024-01-10'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2024-01-10', '2024-01-10'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to date_to', lcm.output)

    @patch('time.time')
    def test_009_validate_dates(self, mock_time):
        """ test validate dates with correct data """
        date_from = '15.01.2024'
        date_to = '10.01.2024'
        mock_time.return_value = 1726309859
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(('2024-01-10', '2024-01-10'), self.validate_dates(date_from, date_to))
        self.assertIn('INFO:dkb_robo.utilities:validate_dates(): adjust date_from to date_to', lcm.output)

    @patch('random.choice')
    def test_010_generate_random_string(self, mock_rc):
        """ test generate_random_string """
        mock_rc.return_value = '1a'
        length = 5
        self.assertEqual('1a1a1a1a1a', self.generate_random_string(length))

    @patch('random.choice')
    def test_011_generate_random_string(self, mock_rc):
        """ generate_random_string """
        mock_rc.return_value = '1a'
        length = 10
        self.assertEqual('1a1a1a1a1a1a1a1a1a1a', self.generate_random_string(length))

    def test_012_string2float(self):
        """ test string2float """
        value = 1000
        self.assertEqual(1000.0, self.string2float(value))

    def test_013_string2float(self):
        """ test string2float """
        value = 1000.0
        self.assertEqual(1000.0, self.string2float(value))

    def test_014_string2float(self):
        """ test string2float """
        value = '1.000,00'
        self.assertEqual(1000.0, self.string2float(value))

    def test_015_string2float(self):
        """ test string2float """
        value = '1000,00'
        self.assertEqual(1000.0, self.string2float(value))

    def test_016_string2float(self):
        """ test string2float """
        value = '1.000'
        self.assertEqual(1000.0, self.string2float(value))

    def test_017_string2float(self):
        """ test string2float """
        value = '1.000,23'
        self.assertEqual(1000.23, self.string2float(value))

    def test_018_string2float(self):
        """ test string2float """
        value = '1000,23'
        self.assertEqual(1000.23, self.string2float(value))

    def test_019_string2float(self):
        """ test string2float """
        value = 1000.23
        self.assertEqual(1000.23, self.string2float(value))

    def test_020_string2float(self):
        """ test string2float """
        value = '-1.000'
        self.assertEqual(-1000.0, self.string2float(value))

    def test_021_string2float(self):
        """ test string2float """
        value = '-1.000,23'
        self.assertEqual(-1000.23, self.string2float(value))

    def test_022_string2float(self):
        """ test string2float """
        value = '-1000,23'
        self.assertEqual(-1000.23, self.string2float(value))

    def test_023_string2float(self):
        """ test string2float """
        value = -1000.23
        self.assertEqual(-1000.23, self.string2float(value))

    def test_024__convert_date_format(self):
        """ test _convert_date_format() """
        self.assertEqual('01.01.2023', self._convert_date_format('2023/01/01', ['%Y/%m/%d'], '%d.%m.%Y'))

    def test_025__convert_date_format(self):
        """ test _convert_date_format() """
        self.assertEqual('wrong date', self._convert_date_format('wrong date', ['%Y/%m/%d'], '%d.%m.%Y'))

    def test_026__convert_date_format(self):
        """ test _convert_date_format() first match """
        self.assertEqual('01.01.2023', self._convert_date_format('2023/01/01', ['%Y/%m/%d', '%d.%m.%Y'], '%d.%m.%Y'))

    def test_027__convert_date_format(self):
        """ test _convert_date_format() last match """
        self.assertEqual('01.01.2023', self._convert_date_format('2023/01/01', ['%d.%m.%Y', '%Y/%m/%d'], '%d.%m.%Y'))

    def test_028__convert_date_format(self):
        """ test _convert_date_format() last match """
        self.assertEqual('2023/01/01', self._convert_date_format('2023/01/01', ['%Y/%m/%d', '%d.%m.%Y'], '%Y/%m/%d'))

    def test_029__convert_date_format(self):
        """ test _convert_date_format() first match """
        self.assertEqual('2023/01/01', self._convert_date_format('2023/01/01', ['%d.%m.%Y', '%Y/%m/%d'], '%Y/%m/%d'))

    def test_030__convert_date_format(self):
        """ test _convert_date_format() no match """
        self.assertEqual('wrong date', self._convert_date_format('wrong date', ['%Y/%m/%d', '%Y-%m-%d'], '%d.%m.%Y'))

    def test_039__get_valid_filename(self):
        """ test get_valid_filename """
        filename = 'test.pdf'
        self.assertEqual('test.pdf', self.get_valid_filename(filename))

    def test_040__get_valid_filename(self):
        """ test get_valid_filename """
        filename = 'test test.pdf'
        self.assertEqual('test_test.pdf', self.get_valid_filename(filename))

    def test_041__get_valid_filename(self):
        """ test get_valid_filename """
        filename = 'testötest.pdf'
        self.assertEqual('testötest.pdf', self.get_valid_filename(filename))

    def test_042__get_valid_filename(self):
        """ test get_valid_filename """
        filename = 'test/test.pdf'
        self.assertEqual('test_test.pdf', self.get_valid_filename(filename))

    def test_043_get_valid_filename(self):
        """ test get_valid_filename """
        filename = 'test\\test.pdf'
        self.assertEqual('test_test.pdf', self.get_valid_filename(filename))

    def test_044_get_valid_filename(self):
        """ test get_valid_filename """
        filename = '.\test.pdf'
        self.assertEqual('._est.pdf', self.get_valid_filename(filename))

    def test_045_get_valid_filename(self):
        """ test get_valid_filename """
        filename = '../test.pdf'
        self.assertEqual('.._test.pdf', self.get_valid_filename(filename))

    @patch('dkb_robo.utilities.generate_random_string')
    def test_046_get_valid_filename(self, mock_rand):
        """ test get_valid_filename """
        filename = '..'
        mock_rand.return_value = 'random'
        self.assertEqual('random.pdf', self.get_valid_filename(filename))


    def test_047_object2dictionary(self):
        """ test object2dictionary """

        nested_obj = DataclassObject(
            Attr1='nested_value1',
            attr2=3,
            attr3='nested_value1',
            attr4='nested_value2'
        )

        test_obj = DataclassObject(
            Attr1='value1',
            attr2=2,
            attr3=nested_obj,
            attr4='foo'
        )

        expected_output = {
            'Attr1': 'value1',
            'attr2': 2,
            'attr3': {'Attr1': 'nested_value1', 'attr2': 3, 'attr3': 'nested_value1', 'attr4': 'nested_value2'},
            'attr4': 'foo'
        }

        result = self.object2dictionary(test_obj)
        self.assertEqual(result, expected_output)


    def test_048_object2dictionary(self):
        """ test object2dictionary """
        test_obj = DataclassObject(
            Attr1='value1',
            attr2=2,
            attr3='attr3',
            attr4='attr4'
        )
        expected_output = {
            'attr1': 'value1',
            'attr3': 'attr3',
            'attr4': 'attr4'
        }
        result = self.object2dictionary(test_obj, key_lc=True, skip_list=['attr2'])
        self.assertEqual(result, expected_output)

    def test_049_logger_setup(self):
        """ logger setup """
        self.assertTrue(self.logger_setup(False))

    def test_050_logger_setup(self):
        """ logger setup """
        self.assertTrue(self.logger_setup(True))

    def test_051_ulal(self):
        """ test ulal() with_valid_parameter """
        mapclass = Mock()
        parameter = {'key1': 'value1', 'key2': 'value2'}
        result = self.ulal(mapclass, parameter)
        mapclass.assert_called_once_with(**parameter)
        self.assertEqual(result, mapclass.return_value)

    def test_052_ulal(self):
        """ test ulal() with none parameter """
        mapclass = MagicMock()
        parameter = None
        result = self.ulal(mapclass, parameter)
        mapclass.assert_not_called()
        self.assertIsNone(result)


class TestAmount(unittest.TestCase):
    """ test class """
    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        from dkb_robo.utilities import Amount
        self.Amount = Amount

    @patch('dkb_robo.utilities.logger', autospec=True)
    def test_047_amount(self, mock_logger):
        """ test Amount """
        amount_data = {
            'value': '1000',
            'currencyCode': 'USD',
            'conversionRate': '1.2',
            'date': '2022-01-01',
            'unit': 'unit'
        }
        amount = self.Amount(**amount_data)
        self.assertEqual(amount.value, 1000.0)
        self.assertEqual(amount.currencyCode, 'USD')
        self.assertEqual(amount.conversionRate, 1.2)
        self.assertEqual(amount.date, '2022-01-01')
        self.assertEqual(amount.unit, 'unit')
        mock_logger.error.assert_not_called()

    @patch('dkb_robo.utilities.logger', autospec=True)
    def test_049_amount(self, mock_logger):
        """ test Amount with wrong value """
        amount_data = {
            'value': 'invalid',
            'currencyCode': 'USD',
            'conversionRate': '1.2',
            'date': '2022-01-01',
            'unit': 'unit'
        }
        amount = self.Amount(**amount_data)
        self.assertFalse(amount.value)
        self.assertEqual(amount.currencyCode, 'USD')
        self.assertEqual(amount.conversionRate, 1.2)
        self.assertEqual(amount.date, '2022-01-01')
        self.assertEqual(amount.unit, 'unit')
        mock_logger.error.assert_called_with('Account.__post_init: value conversion error:  %s', "could not convert string to float: 'invalid'")


    @patch('dkb_robo.utilities.logger', autospec=True)
    def test_050_amount(self, mock_logger):
        """ test Amount with wrong conversation rate """
        amount_data = {
            'value': '1000',
            'currencyCode': 'USD',
            'conversionRate': 'invalid',
            'date': '2022-01-01',
            'unit': 'unit'
        }
        amount = self.Amount(**amount_data)
        self.assertEqual(amount.value, 1000)
        self.assertEqual(amount.currencyCode, 'USD')
        self.assertFalse(amount.conversionRate)
        self.assertEqual(amount.date, '2022-01-01')
        self.assertEqual(amount.unit, 'unit')
        mock_logger.error.assert_called_with('Account.__post_init: converstionRate conversion error:  %s', "could not convert string to float: 'invalid'")


class TestPerformanceValue(unittest.TestCase):
    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        from dkb_robo.utilities import PerformanceValue
        self.PerformanceValue = PerformanceValue


    @patch('dkb_robo.utilities.logger', autospec=True)
    def test_051_performancevalue(self, mock_logger):
        performance_value_data = {
            'currencyCode': 'USD',
            'value': '1000',
            'unit': 'unit'
        }

        performance_value = self.PerformanceValue(**performance_value_data)

        self.assertEqual(performance_value.currencyCode, 'USD')
        self.assertEqual(performance_value.value, 1000.0)
        self.assertEqual(performance_value.unit, 'unit')
        mock_logger.error.assert_not_called()

    @patch('dkb_robo.utilities.logger', autospec=True)
    def test_052_performancevalue(self, mock_logger):
        performance_value_data = {
            'currencyCode': 'USD',
            'value': 'invalid',
            'unit': 'unit'
        }

        performance_value = self.PerformanceValue(**performance_value_data)

        self.assertEqual(performance_value.currencyCode, 'USD')
        self.assertIsNone(performance_value.value)
        self.assertEqual(performance_value.unit, 'unit')
        mock_logger.error.assert_called_with('PerformanceValue.__post_init: conversion error:  %s', "could not convert string to float: 'invalid'")



if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" unittests for dkb_robo """
import sys
import os
import unittest
from unittest.mock import patch, MagicMock, Mock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
from datetime import date
import click
from click.testing import CliRunner

sys.path.insert(0, '.')
sys.path.insert(0, '..')
import logging


class Config():
    def __init__(self):
        self.FORMAT = None
        self.UNFILTERED = False

    def __getitem__(self, key): # this allows getting an element (overrided method)
        return self.FORMAT(key)


class TestDKBRobo(unittest.TestCase):
    """ test class """

    maxDiff = None

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        from dkb_robo.cli import _load_format, _login, standing_orders, credit_limits, last_login, accounts, main, transactions, _id_lookup, _account_lookup, _transactionlink_lookup, scan_postbox, download
        self.logger = logging.getLogger('dkb_robo')
        self._load_format = _load_format
        self._login = _login
        self.standing_orders = standing_orders
        self.credit_limits = credit_limits
        self.last_login = last_login
        self.accounts = accounts
        self.main = main
        self.transactions = transactions
        self._id_lookup = _id_lookup
        self._account_lookup = _account_lookup
        self._transactionlink_lookup = _transactionlink_lookup
        self.scan_postbox = scan_postbox
        self.download = download

    def test_001_default(self):
        """ default test which always passes """
        self.assertEqual('foo', 'foo')


    def test_002__login(self):
        """ test login """
        cursor = MagicMock()
        cursor.__iter__.return_value = []
        self.assertTrue(self._login(cursor))

    def test_003__load_format(self):
        """ test _load_format() """
        oformat = 'pprint'
        self.assertIn('pprint', self._load_format(oformat).__code__.co_names)

    def test_004__load_format(self):
        """ test _load_format() """
        oformat = 'csv'
        self.assertIn('csv', self._load_format(oformat).__code__.co_names)

    def test_005__load_format(self):
        """ test _load_format() """
        oformat = 'table'
        self.assertIn('tabulate', self._load_format(oformat).__code__.co_names)

    def test_006__load_format(self):
        """ test _load_format() """
        oformat = 'json'
        self.assertIn('json', self._load_format(oformat).__code__.co_names)

    def test_007__load_format(self):
        """ test _load_format() """
        oformat = 'foo'
        with self.assertRaises(Exception) as err:
            self._load_format(oformat)
        self.assertEqual('Unknown format: foo', str(err.exception))

    @patch('dkb_robo.cli._account_lookup')
    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_008_standing_orders(self, mock_login, mock_click , mock_lookup):
        """ test standing orders """
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = True
        runner = CliRunner()
        mock_lookup.return_value = 'id'
        self.assertEqual('<Result okay>', str(runner.invoke(self.standing_orders, obj=obj)))
        self.assertFalse(mock_click.called)

    @patch('dkb_robo.cli._id_lookup')
    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_009_standing_orders(self, mock_login, mock_click , mock_lookup):
        """ test standing orders """
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = False
        runner = CliRunner()
        mock_lookup.return_value ='id'
        self.assertEqual('<Result okay>', str(runner.invoke(self.standing_orders, obj=obj)))
        self.assertFalse(mock_click.called)


    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_010_standing_orders(self, mock_login, mock_click):
        """ standing orders """
        from dkb_robo import DKBRoboError
        mock_login.side_effect = DKBRoboError('Error during session confirmation')
        obj = Config()
        obj.FORMAT = 'foo'
        runner = CliRunner()
        self.assertIn('<Result okay>', str(runner.invoke(self.standing_orders, obj=obj)))
        self.assertTrue(mock_click.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_010_credit_limits(self, mock_login, mock_click):
        """ credit limits """
        obj = Config()
        obj.FORMAT = Mock()
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.credit_limits, obj=obj)))
        self.assertFalse(mock_click.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_011_credit_limits(self, mock_login, mock_click):
        """ credit limits """
        from dkb_robo import DKBRoboError
        mock_login.side_effect = DKBRoboError('Error during session confirmation')
        obj = Config()
        obj.FORMAT = 'foo'
        runner = CliRunner()
        self.assertIn('<Result okay>', str(runner.invoke(self.credit_limits, obj=obj)))
        self.assertTrue(mock_click.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_012_last_login(self, mock_login, mock_click):
        """ test last login """
        obj = Config()
        obj.FORMAT = Mock()
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.last_login, obj=obj)))
        self.assertFalse(mock_click.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_013_last_login(self, mock_login, mock_click):
        """ test last login """
        from dkb_robo import DKBRoboError
        mock_login.side_effect = DKBRoboError('Error during session confirmation')
        obj = Config()
        obj.FORMAT = 'foo'
        runner = CliRunner()
        self.assertIn('<Result okay>', str(runner.invoke(self.last_login, obj=obj)))
        self.assertTrue(mock_click.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_014_accounts(self, mock_login, mock_click):
        """ test accounts """
        obj = Config()
        obj.FORMAT = Mock()
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.accounts, obj=obj)))
        self.assertFalse(mock_click.called)

    @patch('dkb_robo.cli.object2dictionary')
    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_015_accounts(self, mock_login, mock_click, mock_object2dictionary):
        """ test accounts """
        mock_login.return_value.__enter__.return_value.account_dic = {1: {'details': 'details', 'transactions': 'transactions'}, 2: {'details': 'details', 'transactions': 'transactions'}}
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = True
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.accounts, obj=obj)))
        self.assertFalse(mock_click.called)
        self.assertTrue(mock_object2dictionary.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_016_accounts(self, mock_login, mock_click):
        """ test accounts """
        from dkb_robo import DKBRoboError
        mock_login.side_effect = DKBRoboError('Error during session confirmation')
        obj = Config()
        obj.FORMAT = 'foo'
        runner = CliRunner()
        self.assertIn('<Result okay>', str(runner.invoke(self.accounts, obj=obj)))
        self.assertTrue(mock_click.called)

    @patch('click.option')
    @patch('click.pass_context')
    def test_017_main(self, mock_pass, mock_option):
        """ test main """

        ctx = MagicMock()
        debug = 'debug'
        use_tan = 'use_tan'
        username = 'username'
        password = 'password'
        format = 'format'
        obj = Config()
        obj.FORMAT = 'foo'

        # self.assertEqual('foo', self.main(ctx, use_tan, username, password, format))
        runner = CliRunner()
        self.assertIn('<Result SystemExit(2)>', str(runner.invoke(self.accounts, [use_tan, username, password, format], obj=obj)))

    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_018_id_lookup(self, mock_account_lookup):
        """ test id look with unfiltered True """
        ctx = MagicMock()
        name = 'test_name'
        account = 'test_account'
        account_dic = {'test_account': {'id': '123'}}
        mock_account_lookup.return_value = MagicMock(id='123')
        result = self._id_lookup(ctx, name, account, account_dic, True)
        mock_account_lookup.assert_called_once_with(ctx, name, account, account_dic, True)
        self.assertEqual(result, '123')

    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_019_id_lookup(self, mock_account_lookup):
        """ test id look with unfiltered False """
        ctx = MagicMock()
        name = 'test_name'
        account = 'test_account'
        account_dic = {'test_account': {'id': '123'}}
        mock_account_lookup.return_value = {'id': '123'}
        result = self._id_lookup(ctx, name, account, account_dic, False)

        mock_account_lookup.assert_called_once_with(ctx, name, account, account_dic, False)
        self.assertEqual(result, '123')

    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_020_id_lookup(self, mock_account_lookup):
        """ test id look with unfiltered False """
        ctx = MagicMock()
        name = 'test_name'
        account = 'test_account'
        account_dic = {'test_account': {'id': '123'}}
        mock_account_lookup.return_value = MagicMock(id=None)
        result = self._id_lookup(ctx, name, account, account_dic, True)
        mock_account_lookup.assert_called_once_with(ctx, name, account, account_dic, True)
        self.assertIsNone(result)

    def test_021_account_lookup(self):
        """ test account lookup by name filtered """
        ctx = MagicMock()
        name = 'Test Account'
        account = None
        account_dic = {
            'acc1': {'name': 'Test Account', 'account': '123456'},
            'acc2': {'name': 'Other Account', 'account': '654321'}
        }
        unfiltered = False
        self.assertEqual(self._account_lookup(ctx, name, account, account_dic, unfiltered), account_dic['acc1'])

    def test_022_account_lookup(self):
        """ test account lookup by name unfiltered """
        ctx = MagicMock()
        name = 'Test Account'
        account = None
        account_dic = {
            'acc1': MagicMock(product=MagicMock(displayName='Test Account')),
            'acc2': MagicMock(product=MagicMock(displayName='Other Account'))
        }
        unfiltered = True
        self.assertEqual( self._account_lookup(ctx, name, account, account_dic, unfiltered), account_dic['acc1'])

    def test_023_account_lookup(self):
        """ test account lookup by account unfiltered """
        ctx = MagicMock()
        name = None
        account = '123456'
        account_dic = {
            'acc1': {'name': 'Test Account', 'account': '123456'},
            'acc2': {'name': 'Other Account', 'account': '654321'}
        }
        unfiltered = False
        result = self._account_lookup(ctx, name, account, account_dic, unfiltered)
        self.assertEqual(result, account_dic['acc1'])

    def test_024_account_lookup(self):
        """ test account lookup by account unfiltered """
        ctx = MagicMock()
        name = None
        account = '123456'
        account_dic = {
            'acc1': MagicMock(type='account', iban='123456'),
            'acc2': MagicMock(type='account', iban='654321')
        }
        unfiltered = True
        self.assertEqual(self._account_lookup(ctx, name, account, account_dic, unfiltered), account_dic['acc1'])

    @patch('dkb_robo.cli.click.echo', autospec=True)
    def test_025_account_lookup(self, mock_echo):
        """ test account lookup no name match"""
        ctx = MagicMock()
        name = 'Nonexistent Account'
        account = None
        account_dic = {
            'acc1': {'name': 'Test Account', 'account': '123456'},
            'acc2': {'name': 'Other Account', 'account': '654321'}
        }
        unfiltered = False
        with self.assertRaises(click.Abort):
            self._account_lookup(ctx, name, account, account_dic, unfiltered)
        mock_echo.assert_called_once_with("No account found matching 'Nonexistent Account'", err=True)

    def test_026_account_lookup(self):
        """ test account lookup neiner name nor account """
        ctx = MagicMock()
        name = None
        account = None
        account_dic = {
            'acc1': {'name': 'Test Account', 'account': '123456'},
            'acc2': {'name': 'Other Account', 'account': '654321'}
        }
        unfiltered = False
        with self.assertRaises(click.UsageError):
            self._account_lookup(ctx, name, account, account_dic, unfiltered)

    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_027_transactionlink_lookup(self, mock_account_lookup):
        """ test transaction link lookup unfiltered True """
        self.ctx = MagicMock()
        name = 'Test Account'
        account = None
        account_dic = {'acc1': MagicMock(id='123', type='account', transactions='transactions_link')}
        unfiltered = True
        mock_account_lookup.return_value = account_dic['acc1']
        result = self._transactionlink_lookup(self.ctx, name, account, account_dic, unfiltered)
        expected_output = {
            'id': '123',
            'type': 'account',
            'transactions': 'transactions_link'
        }
        mock_account_lookup.assert_called_once_with(self.ctx, name, account, account_dic, unfiltered)
        self.assertEqual(result, expected_output)


    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_028_transactionlink_lookup(self, mock_account_lookup):
        """ test transaction link lookup unfiltered False """
        self.ctx = MagicMock()
        name = 'Test Account'
        account = None
        account_dic = {'acc1': {'id': '123', 'type': 'account', 'transactions': 'transactions_link'}}
        unfiltered = False
        mock_account_lookup.return_value = account_dic['acc1']
        result = self._transactionlink_lookup(self.ctx, name, account, account_dic, unfiltered)
        expected_output = {
            'id': '123',
            'type': 'account',
            'transactions': 'transactions_link'
        }
        mock_account_lookup.assert_called_once_with(self.ctx, name, account, account_dic, unfiltered)
        self.assertEqual(result, expected_output)

    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_029_transactionlink_lookup(self, mock_account_lookup):
        """ test transaction link lookup unfiltered True no id """
        self.ctx = MagicMock()
        name = 'Test Account'
        account = None
        account_dic = {'acc1': MagicMock(id=None, type='account', transactions='transactions_link')}
        unfiltered = True
        mock_account_lookup.return_value = account_dic['acc1']
        result = self._transactionlink_lookup(self.ctx, name, account, account_dic, unfiltered)
        expected_output = {
            'id': None,
            'type': 'account',
            'transactions': 'transactions_link'
        }
        mock_account_lookup.assert_called_once_with(self.ctx, name, account, account_dic, unfiltered)
        self.assertEqual(result, expected_output)

    @patch('dkb_robo.cli._account_lookup', autospec=True)
    def test_030_transactionlink_lookup(self, mock_account_lookup):
        """ test transaction link lookup unfiltered False no id """
        self.ctx = MagicMock()
        name = 'Test Account'
        account = None
        account_dic = {'acc1': {'id': None, 'type': 'account', 'transactions': 'transactions_link'}}
        unfiltered = False
        mock_account_lookup.return_value = account_dic['acc1']
        result = self._transactionlink_lookup(self.ctx, name, account, account_dic, unfiltered)
        expected_output = {
            'id': None,
            'type': 'account',
            'transactions': 'transactions_link'
        }
        mock_account_lookup.assert_called_once_with(self.ctx, name, account, account_dic, unfiltered)
        self.assertEqual(result, expected_output)

    @patch('dkb_robo.dkb_robo.DKBRobo.scan_postbox', autospec=True)
    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_031_scan_postbox(self, mock_login, mock_click, mock_scanpb):
        """ test scan postbox """
        mock_login.return_value.__enter__.return_value.account_dic = {1: {'details': 'details', 'transactions': 'transactions'}, 2: {'details': 'details', 'transactions': 'transactions'}}
        mock_scanpb.return_value = {'foo': 'bar'}
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = False
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.scan_postbox, obj=obj)))
        self.assertFalse(mock_click.called)
        # self.assertTrue(mock_scanpb.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_032_scan_postbox(self, mock_login, mock_click):
        """ test scan postbox """
        from dkb_robo import DKBRoboError
        mock_login.side_effect = DKBRoboError('Error during session confirmation')
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = False
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.scan_postbox, obj=obj)))
        self.assertTrue(mock_click.called)


    @patch('dkb_robo.dkb_robo.DKBRobo.scan_postbox', autospec=True)
    @patch('dkb_robo.cli.object2dictionary')
    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_033_scan_postbox(self, mock_login, mock_click, mock_o2d, mock_scanpb):
        """ test scan postbox """
        mock_login.return_value.__enter__.return_value.account_dic = {1: {'details': 'details', 'transactions': 'transactions'}, 2: {'details': 'details', 'transactions': 'transactions'}}
        mock_scanpb.return_value = {'foo': 'bar'}
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = True
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.scan_postbox, obj=obj)))
        self.assertFalse(mock_click.called)
        # self.assertTrue(mock_o2d.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_033_download(self, mock_login, mock_click):
        """ test scan postbox """
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = False
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.download, obj=obj)))
        self.assertFalse(mock_click.called)

    @patch('click.echo')
    @patch('dkb_robo.cli._login')
    def test_034_download(self, mock_login, mock_click):
        """ test scan postbox """
        from dkb_robo import DKBRoboError
        mock_login.side_effect = DKBRoboError('Error during session confirmation')
        obj = Config()
        obj.FORMAT = Mock()
        obj.UNFILTERED = False
        runner = CliRunner()
        self.assertEqual('<Result okay>', str(runner.invoke(self.download, obj=obj)))
        self.assertTrue(mock_click.called)

if __name__ == '__main__':

    unittest.main()


# -*- coding: utf-8 -*-
# pylint: disable=r0904, c0415, c0413, r0913, w0212
""" unittests for dkb_robo """
import sys
import os
from datetime import date
import unittest
import logging
import json
from unittest.mock import patch, Mock, MagicMock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
import io
sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo.exemptionorder import ExemptionOrders

def json_load(fname):
    """ simple json load """

    with open(fname, 'r', encoding='utf8') as myfile:
        data_dic = json.load(myfile)
    return data_dic


class TestExemptionOrders(unittest.TestCase):
    """ test class """

    @patch("requests.Session")
    def setUp(self, mock_session):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.exo = ExemptionOrders(client=mock_session)
        self.maxDiff = None

    @patch('dkb_robo.exemptionorder.ExemptionOrders._filter')
    def test_001_fetch(self, mock_filter):
        """ test ExemptionOrders.fetch() with uid but http error """
        self.exo.client = Mock()
        self.exo.client.get.return_value.status_code = 400
        self.exo.client.get.return_value.json.return_value = {'foo': 'bar'}
        with self.assertRaises(Exception) as err:
            self.assertFalse(self.exo.fetch())
        self.assertEqual('fetch exemption orders: http status code is not 200 but 400', str(err.exception))
        self.assertFalse(mock_filter.called)

    @patch('dkb_robo.exemptionorder.ExemptionOrders._filter')
    def test_002_fetch(self, mock_filter):
        """ test ExemptionOrders.fetch() with uid no error """
        self.exo.client = Mock()
        self.exo.client.get.return_value.status_code = 200
        self.exo.client.get.return_value.json.return_value = {'foo': 'bar'}
        mock_filter.return_value = 'mock_filter'
        self.assertEqual('mock_filter', self.exo.fetch())
        self.assertTrue(mock_filter.called)

    def test_003__filter(self):
        """ test ExemptionOrders._filter() with empty list """
        full_list = {}
        self.assertFalse(self.exo._filter(full_list))

    def test_004__filter(self):
        """ test StandingOrder._filter() with list """
        full_list = {
            'data': {'attributes': {'exemptionCertificates': [],
                         'exemptionOrders': [{'exemptionAmount': {'currencyCode': 'EUR',
                                                                  'value': '2000.00'},
                                              'exemptionOrderType': 'joint',
                                              'partner': {'dateOfBirth': '1970-01-01',
                                                          'firstName': 'Jane',
                                                          'lastName': 'Doe',
                                                          'salutation': 'Frau',
                                                          'taxId': '1234567890'},
                                              'receivedAt': '2020-01-01',
                                              'remainingAmount': {'currencyCode': 'EUR',
                                                                  'value': '1699.55'},
                                              'utilizedAmount': {'currencyCode': 'EUR',
                                                                 'value': '300.50'},
                                              'validFrom': '2020-01-01',
                                              'validUntil': '9999-12-31'}]},
                    'id': 'xxxx',
                    'type': 'customerTaxExemptions'}}

        result = [{'amount': 2000.0, 'used': 300.5, 'currencycode': 'EUR', 'validfrom': '2020-01-01', 'validto': '9999-12-31', 'receivedat': '2020-01-01', 'type': 'joint', 'partner': {'dateofbirth': '1970-01-01', 'firstname': 'Jane', 'lastname': 'Doe', 'salutation': 'Frau', 'taxid': '1234567890'}}]
        self.assertEqual(result, self.exo._filter(full_list))


    def test_005__filter(self):
        """ test StandingOrder._filter() with incomplete list """
        full_list = {
            'data': {'attributes': {'exemptionCertificates': [],
                         'exemptionOrders': [{'exemptionAmount': {'currencyCode': 'EUR',
                                                                  'value': 'aa'},
                                              'exemptionOrderType': 'joint',
                                              'partner': {'dateOfBirth': '1970-01-01',
                                                          'firstName': 'Jane',
                                                          'lastName': 'Doe',
                                                          'salutation': 'Frau',
                                                          'taxId': '1234567890'},
                                              'receivedAt': '2020-01-01',
                                              'remainingAmount': {'currencyCode': 'EUR',
                                                                  'value': 'aa'},
                                              'utilizedAmount': {'currencyCode': 'EUR',
                                                                 'value': 'aa'},
                                              'validFrom': '2020-01-01',
                                              'validUntil': '9999-12-31'}]},
                    'id': 'xxxx',
                    'type': 'customerTaxExemptions'}}

        result = [{'amount': None, 'used': None, 'currencycode': 'EUR', 'validfrom': '2020-01-01', 'validto': '9999-12-31', 'receivedat': '2020-01-01', 'type': 'joint', 'partner': {'dateofbirth': '1970-01-01', 'firstname': 'Jane', 'lastname': 'Doe', 'salutation': 'Frau', 'taxid': '1234567890'}}]
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(result, self.exo._filter(full_list))
        self.assertIn("ERROR:dkb_robo.utilities:Account.__post_init: value conversion error:  could not convert string to float: 'aa'", lcm.output)

    def test_006__filter(self):
        """ test StandingOrder._filter() with list """
        full_list = {
            'data': {'attributes': {'exemptionCertificates': [],
                         'exemptionOrders': [{'exemptionAmount': {'currencyCode': 'EUR',
                                                                  'value': '2000.00'},
                                              'exemptionOrderType': 'joint',
                                              'partner': {'dateOfBirth': '1970-01-01',
                                                          'firstName': 'Jane',
                                                          'lastName': 'Doe',
                                                          'salutation': 'Frau',
                                                          'taxId': '1234567890'},
                                              'receivedAt': '2020-01-01',
                                              'remainingAmount': {'currencyCode': 'EUR',
                                                                  'value': '1699.55'},
                                              'utilizedAmount': {'currencyCode': 'EUR',
                                                                 'value': '300.50'},
                                              'validFrom': '2020-01-01',
                                              'validUntil': '9999-12-31'}]},
                    'id': 'xxxx',
                    'type': 'customerTaxExemptions'}}

        self.exo.unfiltered = True
        result = self.exo._filter(full_list)
        self.assertEqual('2020-01-01', result[0].receivedAt)
        self.assertEqual('EUR', result[0].exemptionAmount.currencyCode)
        self.assertEqual(2000, result[0].exemptionAmount.value)
        self.assertEqual('EUR', result[0].remainingAmount.currencyCode)
        self.assertEqual(1699.55, result[0].remainingAmount.value)
        self.assertEqual('EUR', result[0].utilizedAmount.currencyCode)
        self.assertEqual(300.5, result[0].utilizedAmount.value)
        self.assertEqual('joint', result[0].exemptionOrderType)
        self.assertEqual('2020-01-01', result[0].validFrom)
        self.assertEqual('9999-12-31', result[0].validUntil)
        self.assertEqual('1970-01-01', result[0].partner.dateOfBirth)
        self.assertEqual('Jane', result[0].partner.firstName)
        self.assertEqual('Doe', result[0].partner.lastName)
        self.assertEqual('Frau', result[0].partner.salutation)
        self.assertEqual('1234567890', result[0].partner.taxId)

if __name__ == '__main__':

    unittest.main()



# pylint: disable=c0415, r0913
""" dkb internet banking automation library """
# -*- coding: utf-8 -*-
from pathlib import Path
import time
from dkb_robo.postbox import PostBox
from dkb_robo.authentication import Authentication
from dkb_robo.exemptionorder import ExemptionOrders
from dkb_robo.standingorder import StandingOrders
from dkb_robo.transaction import Transactions
from dkb_robo.utilities import logger_setup, validate_dates, get_dateformat


LEGACY_DATE_FORMAT, API_DATE_FORMAT = get_dateformat()


class DKBRoboError(Exception):
    """ dkb-robo exception class """


class DKBRobo(object):
    """ dkb_robo class """
    # pylint: disable=R0904
    legacy_login = False
    dkb_user = None
    dkb_password = None
    proxies = {}
    last_login = None
    mfa_device = 0
    account_dic = {}
    tan_insert = False
    chip_tan = False
    logger = None
    wrapper = None
    unfiltered = False

    def __init__(self, dkb_user=None, dkb_password=None, tan_insert=False, legacy_login=False, debug=False, mfa_device=None, chip_tan=False, unfiltered=False):
        self.dkb_user = dkb_user
        self.dkb_password = dkb_password
        self.chip_tan = chip_tan
        self.tan_insert = tan_insert
        self.legacy_login = legacy_login
        self.logger = logger_setup(debug)
        self.mfa_device = mfa_device
        self.unfiltered = unfiltered

    def __enter__(self):
        """ Makes DKBRobo a Context Manager """
        # tan usage requires legacy login
        if self.tan_insert:
            self.logger.info('tan_insert is a legacy login option and will be disabled soon. Please use chip_tan instead')
            self.chip_tan = True

        if self.legacy_login:
            raise DKBRoboError('Legacy Login got deprecated. Please do not use this option anymore')

        if self.mfa_device == 'm':
            self.mfa_device = 1

        self.wrapper = Authentication(dkb_user=self.dkb_user, dkb_password=self.dkb_password, proxies=self.proxies, chip_tan=self.chip_tan, mfa_device=self.mfa_device, unfiltered=self.unfiltered)

        # login and get the account overview
        (self.account_dic, self.last_login) = self.wrapper.login()

        return self

    def __exit__(self, *args):
        """ Close the connection at the end of the context """
        self.wrapper.logout()

    def _accounts_by_id(self):
        self.logger.debug('DKBRobo._accounts_by_id()\n')

        if self.unfiltered:
            accounts_by_id = {}
            for acc in self.wrapper.account_dic.values():
                uid = getattr(acc, 'id', None)
                if getattr(acc, 'iban', None):
                    accounts_by_id[uid] = acc.iban
                elif getattr(acc, 'maskedPan', None):
                    accounts_by_id[uid] = acc.maskedPan
                elif getattr(acc, 'depositAccountId', None):
                    accounts_by_id[uid] = acc.depositAccountId
        else:
            accounts_by_id = {acc['id']: acc['account'] for acc in self.wrapper.account_dic.values()}

        self.logger.debug('DKBRobo._accounts_by_id(): returned %s elements\n', len(accounts_by_id.keys()))
        return accounts_by_id

    def get_credit_limits(self):
        """ create a dictionary of credit limits of the different accounts """
        self.logger.debug('DKBRobo.get_credit_limits()\n')

        limit_dic = {}
        for _aid, account_data in self.account_dic.items():
            if 'limit' in account_data:
                if 'iban' in account_data:
                    limit_dic[account_data['iban']] = account_data['limit']
                elif 'maskedpan' in account_data:
                    limit_dic[account_data['maskedpan']] = account_data['limit']

        return limit_dic

    def get_exemption_order(self):
        """ get get_exemption_order """
        self.logger.debug('DKBRobo.get_exemption_order()\n')
        exemptionorder = ExemptionOrders(client=self.wrapper.client, unfiltered=self.unfiltered)
        return exemptionorder.fetch()

    def get_points(self):
        """ returns the DKB points """
        self.logger.debug('DKBRobo.get_points()\n')
        raise DKBRoboError('Method not supported...')

    def get_standing_orders(self, uid=None):
        """ get standing orders """
        self.logger.debug('DKBRobo.get_standing_orders()\n')
        standingorder = StandingOrders(client=self.wrapper.client, unfiltered=self.unfiltered)
        return standingorder.fetch(uid)

    def get_transactions(self, transaction_url, atype, date_from, date_to, transaction_type='booked'):
        """ exported method to get transactions """
        self.logger.debug('DKBRobo.get_transactions(%s/%s: %s/%s)\n', transaction_url, atype, date_from, date_to)

        (date_from, date_to) = validate_dates(date_from, date_to)
        transaction = Transactions(client=self.wrapper.client, unfiltered=self.unfiltered)
        transaction_list = transaction.get(transaction_url, atype, date_from, date_to, transaction_type)

        self.logger.debug('DKBRobo.get_transactions(): %s transactions returned\n', len(transaction_list))
        return transaction_list

    def scan_postbox(self, path=None, download_all=False, _archive=False, prepend_date=False):
        """ scan posbox and return document dictionary """
        self.logger.debug('DKBRobo.scan_postbox()\n')
        return self.download(Path(path) if path is not None else None, download_all, prepend_date)

    def download_doc(self, path: Path, doc, prepend_date: bool = False, mark_read: bool = True, use_account_folders: bool = False, list_only: bool = False, accounts_by_id: dict = None):
        """ download a single document """
        target = path / doc.category()

        if use_account_folders:
            target = target / doc.account(card_lookup=accounts_by_id)

        filename = f"{doc.date()}_{doc.filename()}" if prepend_date else doc.filename()

        if not list_only:
            self.logger.info('Downloading "%s" to %s...', doc.subject(), target)

            download_rcode = doc.download(self.wrapper.client, target / filename)
            if download_rcode:
                if mark_read:
                    doc.mark_read(self.wrapper.client, True)
                time.sleep(.5)
                doc.rcode = download_rcode
            else:
                self.logger.info("File already exists. Skipping %s.", filename)
                doc.rcode = 'skipped'

    def format_doc(self, path, documents, use_account_folders: bool = False, prepend_date: bool = False, accounts_by_id: dict = None):
        """ format documents """

        document_dic = {}
        for doc in documents.values():
            category = doc.category()

            if category not in document_dic:
                document_dic[category] = {
                    'documents': {},
                    'count': 0
                }

            target = path / category
            if use_account_folders:
                target = target / doc.account(card_lookup=accounts_by_id)
            filename = f"{doc.date()}_{doc.filename()}" if prepend_date else doc.filename()

            document_dic[category]['documents'][doc.message.subject] = {
                'id': doc.id,
                'date': doc.document.creationDate,
                'link': doc.document.link,
                'fname': str(target / filename),
                'rcode': doc.rcode

            }
            document_dic[category]['count'] += 1
        return document_dic

    def download(self, path: Path, download_all: bool, prepend_date: bool = False, mark_read: bool = True, use_account_folders: bool = False, list_only: bool = False):
        """ download postbox documents """
        if path is None:
            list_only = True
        postbox = PostBox(client=self.wrapper.client)
        documents = postbox.fetch_items()
        if not download_all:
            # only unread documents
            documents = {id: item for id, item in documents.items()
                         if item.message and item.message.read is False}

        # create dictionary to map accounts to their respective iban/maskedpan
        accounts_by_id = self._accounts_by_id()
        if not list_only:
            # download the documents if required
            for doc in documents.values():
                self.download_doc(path=path, doc=doc, prepend_date=prepend_date, mark_read=mark_read, use_account_folders=use_account_folders, list_only=list_only, accounts_by_id=accounts_by_id)

        # format the documents
        if not self.unfiltered:
            documents = self.format_doc(path=path, documents=documents, use_account_folders=use_account_folders, prepend_date=prepend_date, accounts_by_id=accounts_by_id)

        return documents



""" Module for handling the DKB postbox. """
import datetime
import hashlib
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union
import requests
from dkb_robo.utilities import get_valid_filename, filter_unexpected_fields, DKBRoboError, JSON_CONTENT_TYPE

logger = logging.getLogger(__name__)


@filter_unexpected_fields
@dataclass
class Document:
    """ Document data class, roughly based on the JSON API response. """
    # pylint: disable=c0103
    creationDate: Optional[str] = None
    expirationDate: Optional[str] = None
    retentionPeriod: Optional[str] = None
    contentType: Optional[str] = None
    checksum: Optional[str] = None
    fileName: Optional[str] = None
    metadata: Optional[Union[Dict, str]] = None
    owner: Optional[str] = None
    link: Optional[str] = None
    rcode: Optional[str] = None
    documentTypeId: Optional[str] = None


@filter_unexpected_fields
@dataclass
class Message:
    """ Message data class, roughly based on the JSON API response. """
    # pylint: disable=c0103
    archived: bool = False
    read: bool = False
    subject: Optional[str] = None
    documentId: Optional[str] = None
    documentType: Optional[str] = None
    creationDate: Optional[str] = None
    link: Optional[str] = None


@filter_unexpected_fields
@dataclass
class PostboxItem:
    """ Postbox item data class, merging document and message data and providing download functionality. """
    DOCTYPE_MAPPING = {
        "bankAccountStatement": "Kontoauszüge",
        "creditCardStatement": "Kreditkartenabrechnungen",
        "dwpRevenueStatement": "Wertpapierdokumente",
        "dwpOrderStatement": "Wertpapierdokumente",
        "dwpDepotStatement": "Wertpapierdokumente",
        "exAnteCostInformation": "Wertpapierdokumente",
        "dwpCorporateActionNotice": "Wertpapierdokumente",
    }

    id: str
    document: Document
    message: Message

    def mark_read(self, client: requests.Session, read: bool):
        """ Marks the document as read or unread. """
        logger.debug('PostboxItem.mark_read(): set document %s to %s', self.id, read)
        resp = client.patch(
            self.message.link,
            json={"data": {"attributes": {"read": read}, "type": "message"}},
            headers={"Accept": JSON_CONTENT_TYPE, "Content-type": JSON_CONTENT_TYPE}
        )
        resp.raise_for_status()

    def check_checsum(self, target_file: Path):
        logger.debug('PostboxItem.check_checsum(): %s', self.id)
        with target_file.open('rb') as file:
            if len(self.document.checksum) == 32:
                computed_checksum = hashlib.md5(file.read()).hexdigest()
            elif len(self.document.checksum) == 128:
                computed_checksum = hashlib.sha512(file.read()).hexdigest()
            else:
                raise DKBRoboError(f"Unsupported checksum length: {len(self.document.checksum)}, {self.document.checksum}")
        if computed_checksum != self.document.checksum:
            logger.warning("Checksum mismatch for %s: %s != %s. Renaming file.", target_file, computed_checksum, self.document.checksum)
            # rename file to indicate checksum mismatch
            suffix = '.checksum_mismatch'
            if not target_file.with_name(target_file.name + suffix).exists():
                # rename file to indicate checksum mismatch
                target_file.rename(target_file.with_name(target_file.name + suffix))
            else:
                logger.warning("File %s%s already exists. Not renaming.", target_file, suffix)

    def download(self, client: requests.Session, target_file: Path, overwrite: bool = False):
        """
        Downloads the document from the provided link and saves it to the target file.

        :param client: The requests session to use for downloading the document.
        :param target_file: The path where the document should be saved.
        :param overwrite: Whether to overwrite the file if it already exists.
        :return: True if the file was downloaded and saved, False if the file already exists and overwrite is False.
        """
        logger.debug('PostboxItem.download(): %s to %s', self.id, target_file)
        if not target_file.exists() or overwrite:
            resp = client.get(self.document.link, headers={"Accept": self.document.contentType})
            resp.raise_for_status()

            # create directories if necessary
            target_file.parent.mkdir(parents=True, exist_ok=True)

            with target_file.open('wb') as file:
                file.write(resp.content)

            if self.document.checksum:
                # compare checksums of file with checksum from document metadata
                self.check_checsum(target_file)

            return resp.status_code
        return False

    def filename(self) -> str:
        """ Returns a sanitized filename based on the document metadata. """
        logger.debug('PostboxItem.filename(): Generating filename for document %s', self.id)

        filename = self.document.fileName
        # Depot related files don't have meaningful filenames but only contain the document id. Hence, we use subject
        # instead and rely on the filename sanitization.
        if 'dwpDocumentId' in self.document.metadata and 'subject' in self.document.metadata:
            filename = self.subject() or self.document.fileName

        if self.document.contentType == 'application/pdf' and not filename.endswith('pdf'):
            filename = f'{filename}.pdf'

        fname = get_valid_filename(filename)
        logger.debug('PostboxItem.filename() for %s ended with %s', self.id, fname)
        return fname

    def subject(self) -> str:
        """ Returns the subject of the message. """
        return self.document.metadata.get('subject', self.message.subject)

    def category(self) -> str:
        """ Returns the category of the document based on the document type. """
        return PostboxItem.DOCTYPE_MAPPING.get(self.message.documentType, self.message.documentType)

    def account(self, card_lookup: Dict[str, str] = None) -> str:
        """ Returns the account number or IBAN based on the document metadata. """
        logger.debug("PostboxItem.account() fom document %s", self.id)
        if card_lookup is None:
            card_lookup = {}
        account = None
        if 'depotNumber' in self.document.metadata:
            account = self.document.metadata['depotNumber']
        elif 'cardId' in self.document.metadata:
            account = card_lookup.get(self.document.metadata['cardId'], self.document.metadata['cardId'])
        elif 'iban' in self.document.metadata:
            account = self.document.metadata['iban']

        logger.debug("PostboxItem.account() for document %s ended with %s", self.id, account)
        return account

    def date(self) -> str:
        """ Returns the date of the document based on the metadata. """
        logger.debug("PostboxItem.date() for document %s", self.id)
        date = None
        if 'statementDate' in self.document.metadata:
            date = datetime.date.fromisoformat(self.document.metadata['statementDate'])
        elif 'statementDateTime' in self.document.metadata:
            date = datetime.datetime.fromisoformat(self.document.metadata['statementDateTime'])
        elif 'creationDate' in self.document.metadata:
            date = datetime.date.fromisoformat(self.document.metadata['creationDate'])

        if date is None:
            if 'subject' in self.document.metadata:
                logger.error("\"%s\" is missing a valid date field found in metadata. Using today\'s date as fallback.", self.document.metadata['subject'])
            else:
                logger.error("No valid date field found in document metadata. Using today's date as fallback.")
            date = datetime.date.today()

        logger.debug("PostboxItem.date() for document %s ended with %s", self.id, date)
        return date.strftime('%Y-%m-%d')


class PostBox:
    """ Class for handling the DKB postbox. """
    BASE_URL = "https://banking.dkb.de/api/documentstorage/"

    # pylint: disable=w0621
    def __init__(self, client: requests.Session):
        self.client = client

    def fetch_items(self) -> Dict[str, PostboxItem]:
        """ Fetches all items from the postbox and merges document and message data. """
        logger.debug("PostBox.fetch_items(): Fetching messages")

        def __fix_link_url(url: str) -> str:
            # print(f'old: {url}')
            return url.replace("https://api.developer.dkb.de/documentstorage/", PostBox.BASE_URL)

        response = self.client.get(PostBox.BASE_URL + '/messages')
        response.raise_for_status()
        messages = response.json()

        logger.debug("PostBox.fetch_items(): Fetching documents")
        response = self.client.get(PostBox.BASE_URL + '/documents?page%5Blimit%5D=1000')
        response.raise_for_status()
        documents = response.json()

        if messages and documents:
            # Merge raw messages and documents from JSON API (left join with documents as base).
            items = {
                doc['id']: PostboxItem(
                    id=doc['id'],
                    document=Document(**doc.get('attributes', {}), link=__fix_link_url(doc['links']['self'])),
                    message=None
                )
                for doc in documents.get('data', [])
            }

            # Add matching message data
            for msg in messages.get('data', []):
                msg_id = msg['id']
                if msg_id in items:
                    items[msg_id].message = Message(**msg.get('attributes', {}), link=__fix_link_url(msg['links']['self']))

            return items
        raise DKBRoboError("Could not fetch messages/documents.")



""" Module for handling dkb standing orders """
from typing import Dict, List, Optional, Union
from dataclasses import dataclass, field
import logging
import requests
from dkb_robo.utilities import DKBRoboError, Account, Amount, filter_unexpected_fields, object2dictionary, ulal


logger = logging.getLogger(__name__)


@filter_unexpected_fields
@dataclass
class StandingOrderItem:
    """ class for a single standing order """
    amount: Optional[Dict] = None
    creditor: Optional[Union[Dict, str]] = None
    debtor: Optional[Union[Dict, str]] = None
    description: Optional[str] = None
    messages: List[str] = field(default_factory=list)
    recurrence: Optional[Union[Dict, str]] = None
    status: Optional[str] = None

    def __post_init__(self):
        self.amount = ulal(Amount, self.amount)
        self.creditor['creditorAccount']['name'] = self.creditor.get('name', {})
        self.creditor = ulal(Account, self.creditor['creditorAccount'])
        if self.debtor and 'debtorAccount' in self.debtor:
            self.debtor = ulal(Account, self.debtor['debtorAccount'])
        # rewrite from - field to frm
        self.recurrence['frm'] = self.recurrence.get('from', None)
        self.recurrence = ulal(self.Recurrence, self.recurrence)

    @filter_unexpected_fields
    @dataclass
    class Recurrence:
        """ class for frequency account """
        # pylint: disable=C0103
        frm: Optional[str] = None
        frequency: Optional[str] = None
        holidayExecutionStrategy: Optional[str] = None
        nextExecutionAt: Optional[str] = None
        until: Optional[str] = None


class StandingOrders:
    """ StandingOrders class """
    def __init__(self, client: requests.Session, unfiltered: bool = False, base_url: str = 'https://banking.dkb.de/api'):
        self.client = client
        self.base_url = base_url
        self.unfiltered = unfiltered
        self.uid = None

    def _filter(self, full_list: Dict[str, str]) -> List[Dict[str, str]]:
        """ filter standing orders """
        logger.debug('StandingOrders._filter()\n')

        so_list = []
        if 'data' in full_list:
            for ele in full_list['data']:

                standingorder_obj = StandingOrderItem(**ele['attributes'])

                if self.unfiltered:
                    so_list.append(standingorder_obj)
                else:
                    so_list.append({
                        'amount': standingorder_obj.amount.value,
                        'currencycode': standingorder_obj.amount.currencyCode,
                        'purpose': standingorder_obj.description,
                        'recipient': standingorder_obj.creditor.name,
                        'creditoraccount': object2dictionary(standingorder_obj.creditor, skip_list=['name', 'accountId', 'accountNr', 'id', 'intermediaryName', 'blz']),
                        # from got rewritten in dataclase - we need to rewrite it back
                        'interval': {**object2dictionary(standingorder_obj.recurrence, skip_list=['frm']), 'from': standingorder_obj.recurrence.frm},
                    })

        logger.debug('StandingOrders._filter() ended with: %s entries.', len(so_list))
        return so_list

    def fetch(self, uid) -> Dict:
        """ fetch standing orders """
        logger.debug('StandingOrders.fetch()\n')

        so_list = []
        if uid:
            response = self.client.get(self.base_url + '/accounts/payments/recurring-credit-transfers' + '?accountId=' + uid)
            if response.status_code == 200:
                _so_list = response.json()
                so_list = self._filter(_so_list)
        else:
            raise DKBRoboError('account-id is required to fetch standing orders')

        logger.debug('StandingOrders.fetch() ended\n')
        return so_list



""" miscellaneous functions """
# -*- coding: utf-8 -*-
import logging
from pathlib import Path
import random
from string import digits, ascii_letters
from typing import List, Tuple, Optional
from datetime import datetime, timezone
from dataclasses import dataclass, fields, asdict, is_dataclass
import time
import re


logger = logging.getLogger(__name__)


def get_dateformat():
    """ get date format """
    return '%d.%m.%Y', '%Y-%m-%d'


LEGACY_DATE_FORMAT, API_DATE_FORMAT = get_dateformat()
JSON_CONTENT_TYPE = 'application/vnd.api+json'
BASE_URL = 'https://banking.dkb.de/api'


def filter_unexpected_fields(cls):
    """ filter undefined fields (not defined as class variable) before import to dataclass"""
    original_init = cls.__init__

    def new_init(self, *args, **kwargs):
        expected_fields = {field.name for field in fields(cls)}
        cleaned_kwargs = {key: value for key, value in kwargs.items() if key in expected_fields}
        original_init(self, *args, **cleaned_kwargs)

    cls.__init__ = new_init
    return cls


@filter_unexpected_fields
@dataclass
class Account:
    """ dataclass to build peer account structure """
    # pylint: disable=c0103
    accountNr: Optional[str] = None
    accountId: Optional[str] = None
    bic: Optional[str] = None
    blz: Optional[str] = None
    iban: Optional[str] = None
    id: Optional[str] = None
    intermediaryName: Optional[str] = None
    name: Optional[str] = None


@filter_unexpected_fields
@dataclass
class Amount:
    """ Amount data class, roughly based on the JSON API response. """
    # pylint: disable=c0103
    value: Optional[float] = None
    currencyCode: Optional[str] = None
    conversionRate: Optional[float] = None
    date: Optional[str] = None
    unit: Optional[str] = None

    def __post_init__(self):
        # convert value to float
        try:
            self.value = float(self.value)
        except Exception as err:
            logger.error('Account.__post_init: value conversion error:  %s', str(err))
            self.value = None
        if self.conversionRate:
            try:
                self.conversionRate = float(self.conversionRate)
            except Exception as err:
                logger.error('Account.__post_init: converstionRate conversion error:  %s', str(err))
                self.conversionRate = None


@filter_unexpected_fields
@dataclass
class PerformanceValue:
    """ PerformanceValue data class, roughly based on the JSON API response. """
    # pylint: disable=c0103
    currencyCode: Optional[str] = None
    value: Optional[float] = None
    unit: Optional[str] = None

    def __post_init__(self):
        # convert value to float
        try:
            self.value = float(self.value)
        except Exception as err:
            logger.error('PerformanceValue.__post_init: conversion error:  %s', str(err))
            self.value = None


@filter_unexpected_fields
@dataclass
class Person:
    """ Person class """
    # pylint: disable=c0103
    firstName: Optional[str] = None
    lastName: Optional[str] = None
    title: Optional[str] = None
    salutation: Optional[str] = None
    dateOfBirth: Optional[str] = None
    taxId: Optional[str] = None


class DKBRoboError(Exception):
    """ dkb-robo exception class """


def _convert_date_format(input_date: str, input_format_list: List[str], output_format: str) -> str:
    """ convert date to a specified output format """
    logger.debug('_convert_date_format(%s)', input_date)

    output_date = None
    for input_format in input_format_list:
        try:
            parsed_date = datetime.strptime(input_date, input_format)
            # convert date
            output_date = parsed_date.strftime(output_format)
            break
        except Exception:
            logger.debug('_convert_date_format(): cannot convert date: %s', input_date)
            # something went wrong. we return the date we got as input
            continue

    if not output_date:
        output_date = input_date

    logger.debug('_convert_date_format() ended with: %s', output_date)
    return output_date


def generate_random_string(length: int) -> str:
    """ generate random string to be used as name """
    char_set = digits + ascii_letters
    return ''.join(random.choice(char_set) for _ in range(length))


def get_valid_filename(name):
    """ sanitize filenames """
    s = re.sub(r"(?u)[^-\w.]", " ", str(name))
    p = Path(s.strip())
    s = "_".join(p.stem.split())

    if s in {"", ".", ".."}:
        s = f'{generate_random_string(8)}.pdf'
    return s + p.suffix


def object2dictionary(obj, key_lc=False, skip_list=None):
    """ convert object to dict """

    output_dict = {}
    for k, v in asdict(obj).items():
        if isinstance(skip_list, list) and k in skip_list:
            continue
        if is_dataclass(v):
            output_dict[k] = object2dictionary(v, key_lc=key_lc)
        else:
            if key_lc:
                output_dict[k.lower()] = v
            else:
                output_dict[k] = v
    return output_dict


def string2float(value: str) -> float:
    """ convert string to float value """
    try:
        result = float(value.replace('.', '').replace(',', '.'))
    except Exception:
        result = value

    return result


def logger_setup(debug: bool) -> logging.Logger:
    """ setup logger """
    if debug:
        log_mode = logging.DEBUG
    else:
        log_mode = logging.INFO

    # define standard log format
    log_format = None
    if debug:
        log_format = '%(module)s: %(message)s'
    else:
        log_format = '%(message)s'

    logging.basicConfig(
        format=log_format,
        datefmt="%Y-%m-%d %H:%M:%S",
        level=log_mode)
    mylogger = logging.getLogger('dkb_robo')
    return mylogger


def validate_dates(date_from: str, date_to: str) -> Tuple[str, str]:
    """ correct dates if needed """
    logger.debug('validate_dates()')

    try:
        date_from_uts = int(time.mktime(datetime.strptime(date_from, "%d.%m.%Y").timetuple()))
    except ValueError:
        date_from_uts = int(time.mktime(datetime.strptime(date_from, API_DATE_FORMAT).timetuple()))
    try:
        date_to_uts = int(time.mktime(datetime.strptime(date_to, "%d.%m.%Y").timetuple()))
    except ValueError:
        date_to_uts = int(time.mktime(datetime.strptime(date_to, API_DATE_FORMAT).timetuple()))

    now_uts = int(time.time())

    # ajust valid_from to valid_to
    if date_to_uts <= date_from_uts:
        logger.info('validate_dates(): adjust date_from to date_to')
        date_from = date_to

    # minimal date uts (01.01.2022)
    minimal_date_uts = 1640995200

    if date_from_uts < minimal_date_uts:
        logger.info('validate_dates(): adjust date_from to %s', datetime.fromtimestamp(minimal_date_uts, timezone.utc).strftime(API_DATE_FORMAT))
        date_from = datetime.fromtimestamp(minimal_date_uts, timezone.utc).strftime('%d.%m.%Y')
    if date_to_uts < minimal_date_uts:
        logger.info('validate_dates(): adjust date_to to %s', datetime.fromtimestamp(minimal_date_uts, timezone.utc).strftime(API_DATE_FORMAT))
        date_to = datetime.fromtimestamp(minimal_date_uts, timezone.utc).strftime('%d.%m.%Y')

    if date_from_uts > now_uts:
        logger.info('validate_dates(): adjust date_from to %s', datetime.fromtimestamp(now_uts, timezone.utc).strftime(API_DATE_FORMAT))
        date_from = datetime.fromtimestamp(now_uts).strftime('%d.%m.%Y')
    if date_to_uts > now_uts:
        logger.info('validate_dates(): adjust date_to to %s', datetime.fromtimestamp(now_uts, timezone.utc).strftime(API_DATE_FORMAT))
        date_to = datetime.fromtimestamp(now_uts, timezone.utc).strftime('%d.%m.%Y')

    # this is the new api we need to ensure %Y-%m-%d
    date_from = _convert_date_format(date_from, [API_DATE_FORMAT, LEGACY_DATE_FORMAT], API_DATE_FORMAT)
    date_to = _convert_date_format(date_to, [API_DATE_FORMAT, LEGACY_DATE_FORMAT], API_DATE_FORMAT)

    logger.debug('validate_dates() returned: %s, %s', date_from, date_to)
    return date_from, date_to


def ulal(mapclass, parameter):
    """ map parameter """
    if parameter:
        return mapclass(**parameter)
    return None



# -*- coding: utf-8 -*-
# pylint: disable=r0904, c0415, c0413, r0913, w0212
""" unittests for dkb_robo """
import sys
import os
from datetime import date
import unittest
import logging
import json
from unittest.mock import patch, Mock, MagicMock, mock_open
from bs4 import BeautifulSoup
from mechanicalsoup import LinkNotFoundError
import io
sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo.standingorder import StandingOrders

def json_load(fname):
    """ simple json load """

    with open(fname, 'r', encoding='utf8') as myfile:
        data_dic = json.load(myfile)

    return data_dic


class TestDKBRobo(unittest.TestCase):
    """ test class """

    @patch("requests.Session")
    def setUp(self, mock_session):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.logger = logging.getLogger('dkb_robo')
        self.dkb = StandingOrders(client=mock_session)
        self.maxDiff = None

    @patch('dkb_robo.standingorder.StandingOrders._filter')
    def test_001_fetch(self, mock_filter):
        """ test StandingOrders.fetch() without uid """
        self.dkb.client = Mock()
        self.dkb.client.get.return_value.status_code = 200
        self.dkb.client.get.return_value.json.return_value = {'foo': 'bar'}

        with self.assertRaises(Exception) as err:
            self.assertFalse(self.dkb.fetch(None))
        self.assertEqual('account-id is required to fetch standing orders', str(err.exception))
        self.assertFalse(mock_filter.called)

    @patch('dkb_robo.standingorder.StandingOrders._filter')
    def test_002_fetch(self, mock_filter):
        """ test StandingOrders.fetch() with uid but http error """
        self.dkb.client = Mock()
        self.dkb.client.get.return_value.status_code = 400
        self.dkb.client.get.return_value.json.return_value = {'foo': 'bar'}
        self.assertFalse(self.dkb.fetch(uid='uid'))
        self.assertFalse(mock_filter.called)

    @patch('dkb_robo.standingorder.StandingOrders._filter')
    def test_003_fetch(self, mock_filter):
        """ test StandingOrders.fetch() with uid no error """
        self.dkb.client = Mock()
        self.dkb.client.get.return_value.status_code = 200
        self.dkb.client.get.return_value.json.return_value = {'foo': 'bar'}
        mock_filter.return_value = 'mock_filter'
        self.assertEqual('mock_filter', self.dkb.fetch(uid='uid'))
        self.assertTrue(mock_filter.called)

    def test_004__filter(self):
        """ test StandingOrders._filter() with empty list """
        full_list = {}
        self.assertFalse(self.dkb._filter(full_list))

    def test_005__filter(self):
        """ test StandingOrders._filter() with list """
        full_list = {
            "data": [
                {
                    "attributes": {
                        "description": "description",
                        "amount": {
                            "currencyCode": "EUR",
                            "value": "100.00"
                        },
                        "creditor": {
                            "name": "cardname",
                            "creditorAccount": {
                                "iban": "crediban",
                                "bic": "credbic"
                            }
                        },
                        "recurrence": {
                            "from": "2020-01-01",
                            "until": "2025-12-01",
                            "frequency": "monthly",
                            "nextExecutionAt": "2020-02-01"
                        }
                    }
                }]}
        result = [{'amount': 100.0, 'currencycode': 'EUR', 'purpose': 'description', 'recipient': 'cardname', 'creditoraccount': {'iban': 'crediban', 'bic': 'credbic'}, 'interval': {'from': '2020-01-01', 'until': '2025-12-01', 'frequency': 'monthly', 'nextExecutionAt': '2020-02-01', 'holidayExecutionStrategy': None}}]
        self.assertEqual(result, self.dkb._filter(full_list))

    def test_006__filter(self):
        """ test StandingOrders._filter() with list from file """
        so_list = json_load(self.dir_path + '/mocks/so.json')
        self.dkb.unfiltered = False
        result = [{'amount': 100.0, 'currencycode': 'EUR', 'purpose': 'description1', 'recipient': 'name1', 'creditoraccount': {'iban': 'iban1', 'bic': 'bic1'}, 'interval': {'from': '2022-01-01', 'until': '2025-12-01', 'frequency': 'monthly', 'holidayExecutionStrategy': 'following', 'nextExecutionAt': '2022-11-01'}}, {'amount': 200.0, 'currencycode': 'EUR', 'purpose': 'description2', 'recipient': 'name2', 'creditoraccount': {'iban': 'iban2', 'bic': 'bic2'}, 'interval': {'from': '2022-02-01', 'until': '2025-12-02', 'frequency': 'monthly', 'holidayExecutionStrategy': 'following', 'nextExecutionAt': '2022-11-02'}}, {'amount': 300.0, 'currencycode': 'EUR', 'purpose': 'description3', 'recipient': 'name3', 'creditoraccount': {'iban': 'iban3', 'bic': 'bic3'}, 'interval': {'from': '2022-03-01', 'until': '2025-03-01', 'frequency': 'monthly', 'holidayExecutionStrategy': 'following', 'nextExecutionAt': '2022-03-01'}}]
        self.assertEqual(result, self.dkb._filter(so_list))

    def test_007__filter(self):
        """ test StandingOrders._filter() with incomplete list/conversion error """
        full_list = {
            "data": [
                {
                    "attributes": {
                        "description": "description",
                        "amount": {
                            "value": "aa"
                        },
                        "creditor": {
                            "name": "cardname",
                            "creditorAccount": {
                                "iban": "crediban",
                                "bic": "credbic"
                            }
                        },
                        "recurrence": {
                            "from": "2020-01-01",
                            "until": "2025-12-01",
                            "frequency": "monthly",
                            "nextExecutionAt": "2020-02-01"
                        }
                    }
                }]}
        result = [{'amount': None, 'currencycode': None, 'purpose': 'description', 'recipient': 'cardname', 'creditoraccount': {'iban': 'crediban', 'bic': 'credbic'}, 'interval': {'from': '2020-01-01', 'until': '2025-12-01', 'frequency': 'monthly', 'nextExecutionAt': '2020-02-01', 'holidayExecutionStrategy': None}}]
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual(result, self.dkb._filter(full_list))
        self.assertIn("ERROR:dkb_robo.utilities:Account.__post_init: value conversion error:  could not convert string to float: 'aa'", lcm.output)

    def test_008__filter(self):
        """ test StandingOrders._filter() with incomplete list/conversion error """
        full_list = {
            "data": [
                {
                    "attributes": {
                        "description": "description",
                        "amount": {
                            "value": "100"
                        },
                        "creditor": {
                            "name": "cardname",
                            "creditorAccount": {
                                "iban": "crediban",
                                "bic": "credbic"
                            }
                        },
                        "recurrence": {
                            "from": "2020-01-01",
                            "until": "2025-12-01",
                            "frequency": "monthly",
                            "nextExecutionAt": "2020-02-01"
                        }
                    }
                }]}
        self.dkb.unfiltered = True
        result = self.dkb._filter(full_list)
        self.assertEqual(100, result[0].amount.value)
        self.assertEqual('description', result[0].description)
        self.assertEqual('cardname', result[0].creditor.name)
        self.assertEqual('crediban', result[0].creditor.iban)
        self.assertEqual('credbic', result[0].creditor.bic)
        self.assertEqual('2020-01-01', result[0].recurrence.frm)
        self.assertEqual('monthly', result[0].recurrence.frequency)
        self.assertEqual('2025-12-01', result[0].recurrence.until)
        self.assertEqual('2020-02-01', result[0].recurrence.nextExecutionAt)
        self.assertIsNone(result[0].recurrence.holidayExecutionStrategy)

if __name__ == '__main__':

    unittest.main()



"""Tests for the postbox module."""
import sys
import tempfile
import unittest
from unittest.mock import patch, MagicMock
from pathlib import Path
from datetime import date
import requests
import logging
sys.path.insert(0, '.')
sys.path.insert(0, '..')
from dkb_robo.utilities import DKBRoboError
from dkb_robo.postbox import PostboxItem, PostBox, Document, Message


class TestPostboxItem(unittest.TestCase):
    """ Tests for the PostboxItem class. """

    def setUp(self):
        self.document = Document(
            creationDate="2023-01-01",
            expirationDate="2023-12-31",
            retentionPeriod="9999-12-31",
            contentType="application/pdf",
            checksum="9473fdd0d880a43c21b7778d34872157",
            fileName="test_document",
            metadata={"statementDate": "2023-01-01"},
            owner="owner",
            link="http://example.com/document",
        )
        self.message = Message(
            archived=False,
            read=False,
            subject="Test Subject",
            documentType="bankAccountStatement",
            creationDate="2023-01-01",
            link="http://example.com/message",
        )
        self.postbox_item = PostboxItem(
            id="1", document=self.document, message=self.message,
        )

    @patch("requests.Session")
    def test_001_mark_read(self, mock_session):
        """Test that the mark_read method correctly marks a document as read or unread."""
        mock_client = mock_session.return_value
        mock_client.patch.return_value.status_code = 200
        self.postbox_item.mark_read(mock_client, True)
        mock_client.patch.assert_called_once_with(
            self.message.link,
            json={"data": {"attributes": {"read": True}, "type": "message"}},
            headers={
                "Accept": "application/vnd.api+json",
                "Content-type": "application/vnd.api+json",
            },
        )

    @patch("requests.Session")
    def test_002_mark_read_failed(self, mock_session):
        """ Test that the mark_read method raises an exception if the request fails. """
        mock_client = mock_session.return_value
        mock_client.patch.return_value.raise_for_status.side_effect = (
            requests.HTTPError()
        )
        with self.assertRaises(requests.HTTPError):
            self.postbox_item.mark_read(mock_client, True)

    @patch("requests.Session")
    def test_003_download(self, mock_session):
        """ Test that the download method correctly downloads a document and saves it to the specified file path."""
        mock_client = mock_session.return_value
        mock_client.get.return_value.status_code = 200
        mock_client.get.return_value.content = b"test content"
        target_file = Path(tempfile.gettempdir()) / "test_001_document.pdf"
        result = self.postbox_item.download(mock_client, target_file, overwrite=True)
        self.assertTrue(result)
        self.assertTrue(target_file.exists())
        target_file.unlink()  # Remove the downloaded test file

    @patch("requests.Session")
    def test_004_download_existing_file(self, mock_session):
        """ Test that the download method does not overwrite an existing file if overwrite is False."""
        mock_client = mock_session.return_value
        target_file = Path(tempfile.gettempdir()) / "existing.pdf"
        target_file.touch()
        result = self.postbox_item.download(mock_client, target_file)
        self.assertFalse(result)
        target_file.unlink()

    @patch("requests.Session")
    def test_005_download_checksum_mismatch(self, mock_session):
        """ Test that the download method renames the downloaded file if the checksum of the downloaded file does not match."""
        mock_client = mock_session.return_value
        mock_client.get.return_value.status_code = 200
        mock_client.get.return_value.content = b"wrong test content"
        target_file = Path(tempfile.gettempdir()) / "test_document.pdf"
        result = self.postbox_item.download(mock_client, target_file, overwrite=True)
        self.assertTrue(result)
        self.assertFalse(target_file.exists())
        mismatched_file = target_file.with_name(target_file.name + ".checksum_mismatch")
        self.assertTrue(mismatched_file.exists())
        mismatched_file.unlink()  # Remove the downloaded test file

    @patch("pathlib.Path.exists")
    @patch("requests.Session")
    def test_006_download_file_exists_no_overwrite(self, mock_session, mock_exists):
        """ Test that the download method renames the downloaded file if the checksum of the downloaded file does not match."""
        mock_client = mock_session.return_value
        mock_client.get.return_value.status_code = 200
        mock_client.get.return_value.content = b"wrong test content"
        target_file = Path("document.pdf")
        mock_exists.side_effect = [False, True, True]
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            result = self.postbox_item.download(mock_client, target_file, overwrite=True)
        self.assertIn('WARNING:dkb_robo.postbox:File document.pdf.checksum_mismatch already exists. Not renaming.', lcm.output)
        self.assertTrue(result)
        target_file.unlink()  # Remove the downloaded test file

    def test_007_filename(self):
        """ Test that the filename method returns the correct filename for the postbox item."""
        filename = self.postbox_item.filename()
        self.assertEqual(filename, "test_document.pdf")

    def test_008_category(self):
        """ Test that the category method returns the correct category for the postbox item."""
        category = self.postbox_item.category()
        self.assertEqual(category, "Kontoauszüge")

    def test_009_account(self):
        """ Test that the account method returns the correct account for the postbox item."""
        account = self.postbox_item.account()
        self.assertIsNone(account)

    def test_010_date(self):
        """ Test that the date method returns the correct date for the postbox item."""
        date = self.postbox_item.date()
        self.assertEqual(date, "2023-01-01")

    def test_011_filename_with_depot_document(self):
        """ Test that the filename method returns the correct filename for a depot document."""
        self.document.metadata = {"dwpDocumentId": "123", "subject": "Depot Statement"}
        filename = self.postbox_item.filename()
        self.assertEqual(filename, "Depot_Statement.pdf")

    def test_012_filename_basic(self):
        """ Test basic filename return without modification """
        self.document.fileName = "simple_document.pdf"
        self.assertEqual(self.postbox_item.filename(), "simple_document.pdf")

    def test_013_filename_adds_pdf_extension(self):
        """ Test that .pdf extension is added for PDF content type """
        self.document.fileName = "document_without_extension"
        self.document.contentType = "application/pdf"
        self.assertEqual(self.postbox_item.filename(), "document_without_extension.pdf")

    def test_014_filename_sanitization(self):
        """ Test filename sanitization with special characters """
        self.document.fileName = "Document Ümlaut & Spaces!.pdf"
        self.assertEqual(self.postbox_item.filename(), "Document_Ümlaut_Spaces.pdf")

    def test_015_filename_multiple_whitespace(self):
        """ Test handling of multiple whitespaces in filename """
        self.document.fileName = "12345.pdf"
        self.document.metadata = {
            "dwpDocumentId": "12345",
            "subject": "Depot    Statement     Q1    2023"
        }
        self.assertEqual(self.postbox_item.filename(), "Depot_Statement_Q1_2023.pdf")

    def test_016_filename_empty_subject(self):
        """ Test handling of empty subject in depot documents """
        self.document.fileName = "fallback.pdf"
        self.document.metadata = {
            "dwpDocumentId": "12345",
            "subject": ""
        }
        self.assertEqual(self.postbox_item.filename(), "fallback.pdf")

    def test_017_account_with_depot(self):
        """ Test that the account method returns the correct account for a depot document."""
        self.document.metadata = {"depotNumber": "12345"}
        self.assertEqual(self.postbox_item.account(), "12345")

    def test_018_account_with_card(self):
        """ Test that the account method returns the correct account for a credit card document."""
        self.document.metadata = {"cardId": "card123"}
        card_lookup = {"card123": "MyCard"}
        self.assertEqual(self.postbox_item.account(card_lookup), "MyCard")

    def test_019_account_with_card_failed_lookup(self):
        """ Test that the account method returns the card ID if the lookup fails."""
        self.document.metadata = {"cardId": "card123"}
        card_lookup = {"card1234": "MyCard"}
        self.assertEqual(self.postbox_item.account(card_lookup), "card123")

    def test_020_account_with_iban(self):
        """ Test that the account method returns the correct account for an IBAN document."""
        self.document.metadata = {"iban": "DE123"}
        self.assertEqual(self.postbox_item.account(), "DE123")

    def test_021_date_with_statement_datetime(self):
        """ Test that the date method returns the correct date for a document with a statementDateTime field."""
        self.document.metadata = {"statementDateTime": "2023-01-01T12:00:00"}
        self.assertEqual(self.postbox_item.date(), "2023-01-01")

    def test_022_date_with_statement_date(self):
        """ Test that the date method returns the correct date for a document with a statementDate field."""
        self.document.metadata = {"statementDate": "2023-01-01"}
        self.assertEqual(self.postbox_item.date(), "2023-01-01")

    def test_023_date_with_creation_date(self):
        """ Test that the date method returns the correct date for a document with a creationDate field."""
        self.document.metadata = {"creationDate": "2023-01-01"}
        self.assertEqual(self.postbox_item.date(), "2023-01-01")

    @patch("dkb_robo.postbox.datetime.date")
    def test_024_date_invalid(self, mock_today):
        """ Test that the date method raises an exception if no valid date field is found in the metadata."""
        self.document.metadata = {}
        mock_today.today.return_value = date(2025, 3, 23)
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual('2025-03-23', self.postbox_item.date())
        self.assertIn("ERROR:dkb_robo.postbox:No valid date field found in document metadata. Using today's date as fallback.", lcm.output)

    @patch("dkb_robo.postbox.datetime.date")
    def test_025_date_invalid(self, mock_today):
        """ Test that the date method raises an exception if no valid date field is found in the metadata."""
        self.document.metadata = {"subject": "subject"}
        mock_today.today.return_value = date(2025, 3, 23)
        with self.assertLogs('dkb_robo', level='INFO') as lcm:
            self.assertEqual('2025-03-23', self.postbox_item.date())
        self.assertIn('ERROR:dkb_robo.postbox:"subject" is missing a valid date field found in metadata. Using today\'s date as fallback.', lcm.output)

    @patch("requests.Session")
    def test_031_download_sha512_checksum(self, mock_session):
        """ Test that the download method correctly handles SHA-512 checksums. """
        # Create a SHA-512 checksum (128 characters long)
        sha512_checksum = "a7f06f12bd5e99c59966c13cea8483ece69d27a3bf8d46f836e934f5c1ec0b55a7e22d14360949de102cd8e16a9f74806c9ce1597eefaa379fb40384403c7fd6"
        self.document.checksum = sha512_checksum

        # Mock the session and content
        mock_client = mock_session.return_value
        mock_client.get.return_value.status_code = 200
        content = b"test content for SHA-512 verification"
        mock_client.get.return_value.content = content

        # Create a target file
        target_file = Path(tempfile.gettempdir()) / "test_sha512_document.pdf"

        # Mock the SHA-512 hash computation
        with patch("hashlib.sha512") as mock_sha512:
            mock_sha512.return_value.hexdigest.return_value = sha512_checksum

            # Test the download
            result = self.postbox_item.download(mock_client, target_file, overwrite=True)

            # Verify the results
            self.assertTrue(result)
            self.assertTrue(target_file.exists())
            mock_sha512.assert_called_once_with(content)
        target_file.unlink()  # Remove the downloaded test file

    @patch("requests.Session")
    def test_032_download_sha512_checksum_mismatch(self, mock_session):
        """ Test that the download method handles SHA-512 checksum mismatches correctly. """
        # Create a SHA-512 checksum (128 characters long)
        sha512_checksum = "a7f06f12bd5e99c59966c13cea8483ece69d27a3bf8d46f836e934f5c1ec0b55a7e22d14360949de102cd8e16a9f74806c9ce1597eefaa379fb40384403c7fd6"
        self.document.checksum = sha512_checksum

        # Mock the session and content
        mock_client = mock_session.return_value
        mock_client.get.return_value.status_code = 200
        content = b"test content that will cause a SHA-512 checksum mismatch"
        mock_client.get.return_value.content = content

        # Create a target file
        target_file = Path(tempfile.gettempdir()) / "test_sha512_mismatch.pdf"

        # Set up a different checksum to be returned by the SHA-512 hash function to simulate a mismatch
        wrong_checksum = "b8f06f12bd5e99c59966c13cea8483ece69d27a3bf8d46f836e934f5c1ec0b55a7e22d14360949de102cd8e16a9f74806c9ce1597eefaa379fb40384403c7fd7"

        with patch("hashlib.sha512") as mock_sha512:
            mock_sha512.return_value.hexdigest.return_value = wrong_checksum

            # Test the download
            result = self.postbox_item.download(mock_client, target_file, overwrite=True)

            # Verify the results
            self.assertTrue(result)
            self.assertFalse(target_file.exists())
            mismatched_file = target_file.with_name(target_file.name + ".checksum_mismatch")
            self.assertTrue(mismatched_file.exists())
            mismatched_file.unlink()  # Remove the downloaded test file

    @patch("requests.Session")
    def test_033_download_unsupported_checksum_length(self, mock_session):
        """ Test that the download method raises an error for unsupported checksum lengths. """
        # Create an unsupported checksum length (e.g., 64 characters)
        unsupported_checksum = "a" * 64
        self.document.checksum = unsupported_checksum

        # Mock the session and content
        mock_client = mock_session.return_value
        mock_client.get.return_value.status_code = 200
        mock_client.get.return_value.content = b"test content"

        # Create a target file
        target_file = Path(tempfile.gettempdir()) / "test_unsupported_checksum.pdf"

        # Test the download function and expect a DKBRoboError
        with self.assertRaises(DKBRoboError) as context:
            self.postbox_item.download(mock_client, target_file, overwrite=True)

        # Verify the error message
        self.assertIn(f"Unsupported checksum length: {len(unsupported_checksum)}", str(context.exception))
        target_file.unlink()  # Remove the downloaded test file

class TestPostBox(unittest.TestCase):
    """ Tests for the PostBox class. """

    def create_mocked_postbox(self, mock_client):
        """ Create a PostBox instance with a mocked client. """
        mock_client.get.side_effect = [
            MagicMock(
                status_code=200,
                json=lambda: {
                    "data": [
                        {
                            "id": "11111111-1111-1111-1111-111111111111",
                            "attributes": {
                                "read": "True",
                                "archived": "False",
                                "subject": "Kreditkartenabrechnung vom 22. September 2023",
                                "documentType": "creditCardStatement",
                                "creationDate": "2023-09-26T06:57:34.803691Z",
                            },
                            "links": {
                                "self": "https://api.developer.dkb.de/messages/11111111-1111-1111-1111-111111111111"
                            },
                        }
                    ]
                },
            ),
            MagicMock(
                status_code=200,
                json=lambda: {
                    "data": [
                        {
                            "id": "11111111-1111-1111-1111-111111111111",
                            "attributes": {
                                "creationDate": "2023-09-01T01:01:03.456789Z",
                                "expirationDate": "2026-09-01",
                                "retentionPeriod": "P3Y",
                                "contentType": "application/pdf",
                                "checksum": "00000000000000000000000000000000",
                                "fileName": "Kreditkarte_XXXXXXXXXXXXXXXX_Abrechnung_20230901.pdf",
                                "metadata": {
                                    "statementAmount": "47.11",
                                    "statementDate": "2023-09-01",
                                    "subject": "Kreditkartenabrechnung vom 01. September 2023",
                                    "statementID": "_X_XXXXXX",
                                    "statementCurrency": "EUR",
                                    "portfolio": "dkb",
                                    "cardId": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
                                },
                                "owner": "service-documentexchange-api",
                            },
                            "links": {
                                "self": "https://api.developer.dkb.de/documentstorage/documents/11111111-1111-1111-1111-111111111111"
                            },
                        }
                    ]
                },
            ),
        ]
        return PostBox(client=mock_client)

    @patch("requests.Session")
    def test_026_fetch_items(self, mock_session):
        """ Test that the fetch_items method correctly fetches items from the postbox. """
        postbox = self.create_mocked_postbox(mock_session.return_value)
        items = postbox.fetch_items()
        self.assertIn("11111111-1111-1111-1111-111111111111", items)
        self.assertIsInstance(
            items["11111111-1111-1111-1111-111111111111"], PostboxItem
        )

    @patch("requests.Session")
    def test_027_sample(self, mock_session):
        """ Test that the fetch_items method correctly fetches items from the postbox. """
        item = self.create_mocked_postbox(mock_session.return_value).fetch_items()[
            "11111111-1111-1111-1111-111111111111"
        ]

        self.assertEqual(
            item.filename(), "Kreditkarte_XXXXXXXXXXXXXXXX_Abrechnung_20230901.pdf"
        )
        self.assertEqual(item.account(), "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee")
        self.assertEqual(item.date(), "2023-09-01")
        self.assertEqual(item.category(), "Kreditkartenabrechnungen")

    @patch("requests.Session")
    def test_028_fetch_empty_responses(self, mock_session):
        """ Test that the fetch_items method raises an exception if the responses are empty. """
        mock_client = mock_session.return_value
        mock_client.get.side_effect = [
            MagicMock(status_code=200, json=lambda: {}),
            MagicMock(status_code=200, json=lambda: {}),
        ]
        with self.assertRaises(DKBRoboError):
            PostBox(client=mock_client).fetch_items()

    @patch("requests.Session")
    def test_029_fetch_url_fixing(self, mock_session):
        """ Test that the link URLs are correctly fixed."""
        postbox = self.create_mocked_postbox(mock_session.return_value)
        self.assertTrue(
            postbox.fetch_items()[
                "11111111-1111-1111-1111-111111111111"
            ].document.link.startswith(PostBox.BASE_URL)
        )

    @patch("requests.Session")
    def test_030_fetch_http_error(self, mock_session):
        """ Test that the fetch_items method raises an exception if the request fails. """
        mock_client = mock_session.return_value
        mock_client.get.side_effect = requests.HTTPError()
        with self.assertRaises(requests.HTTPError):
            PostBox(client=mock_client).fetch_items()


if __name__ == "__main__":
    unittest.main()



# pylint: disable=r0913
""" Module for handling dkb standing orders """
from typing import Dict, List, Tuple
import time
import json
import base64
import io
import threading
import logging
import requests
from dkb_robo.legacy import Wrapper as Legacywrapper
from dkb_robo.portfolio import Overview
from dkb_robo.utilities import DKBRoboError, JSON_CONTENT_TYPE


BASE_URL = 'https://banking.dkb.de/api'
logger = logging.getLogger(__name__)


class Authentication:
    """ Authentication class """

    account_dic = {}
    base_url = BASE_URL
    chip_tan = False
    client = None
    dkb_user = None
    dkb_password = None
    dkb_br = None
    logger = None
    mfa_method = 'seal_one'
    mfa_device = 0
    proxies = {}
    token_dic = None
    unfiltered = False

    def __init__(self, dkb_user: str = None, dkb_password: str = None, chip_tan: bool = False, proxies: Dict[str, str] = None, mfa_device: int = None, unfiltered: bool = False):
        """ Constructor """
        self.chip_tan = chip_tan
        self.dkb_user = dkb_user
        self.dkb_password = dkb_password
        self.proxies = proxies
        self.unfiltered = unfiltered
        if chip_tan:
            logger.info('Using to chip_tan to login')
            if chip_tan in ('qr', 'chip_tan_qr'):
                self.mfa_method = 'chip_tan_qr'
            else:
                self.mfa_method = 'chip_tan_manual'
        try:
            self.mfa_device = int(mfa_device)
        except (ValueError, TypeError):
            self.mfa_device = 0

    def _mfa_challenge(self, mfa_dic: Dict[str, str], device_num: int = 0) -> Tuple[str, str]:
        """ get challenge dict with information on the 2nd factor """
        logger.debug('Authentication._mfa_challenge(): login with device_num: %s\n', device_num)

        device_name = None
        challenge_dic = {}
        if 'data' in mfa_dic and 'id' in mfa_dic['data'][device_num]:
            try:
                device_name = mfa_dic['data'][device_num]['attributes']['deviceName']
                logger.debug('api.Wrapper._mfa_challenge(): devicename: %s\n', device_name)
            except Exception as _err:
                logger.error('unable to get mfa-deviceName for device_num: %s', device_num)
                device_name = None

            # additional headers needed as this call requires it
            self.client.headers['Content-Type'] = JSON_CONTENT_TYPE
            self.client.headers["Accept"] = "application/vnd.api+json"

            # we are expecting the first method from mfa_dic to be used
            data_dic = {'data': {'type': 'mfa-challenge', 'attributes': {'mfaId': self.token_dic['mfa_id'], 'methodId': mfa_dic['data'][device_num]['id'], 'methodType': self.mfa_method}}}
            response = self.client.post(self.base_url + '/mfa/mfa/challenges', data=json.dumps(data_dic))

            # process response
            # challenge_id = self._process_challenge_response(response)
            if response.status_code in (200, 201):
                challenge_dic = response.json()
            else:
                raise DKBRoboError(f'Login failed: post request to get the mfa challenges failed. RC: {response.status_code}')

            # we rmove the headers we added earlier
            self.client.headers.pop('Content-Type')
            self.client.headers.pop('Accept')
        else:
            logger.error('mfa dictionary has an unexpected data structure')

        logger.debug('Authentication._mfa_challenge() ended\n')
        return challenge_dic, device_name

    def _mfa_challenge_id(self, challenge_dic: Dict[str, str]) -> str:
        """ get challenge dict with information on the 2nd factor """
        logger.debug('api.Wrapper._mfa_challenge_id()\n')
        challenge_id = None

        if 'data' in challenge_dic and 'id' in challenge_dic['data'] and 'type' in challenge_dic['data']:
            if challenge_dic['data']['type'] == 'mfa-challenge':
                challenge_id = challenge_dic['data']['id']
            else:
                raise DKBRoboError(f'Login failed:: wrong challenge type: {challenge_dic}')

        else:
            raise DKBRoboError(f'Login failed: challenge response format is other than expected: {challenge_dic}')

        logger.debug('api.Wrapper._mfa_challenge_id() ended with: %s\n', challenge_id)
        return challenge_id

    def _mfa_finalize(self, challenge_dic: Dict[str, str], devicename: str) -> bool:
        """ wait for confirmation for the 2nd factor """
        logger.debug('Authentication._mfa_finalize()\n')

        challenge_id = self._mfa_challenge_id(challenge_dic)

        if self.mfa_method == 'seal_one':
            mfa_auth = APPAuthentication(client=self.client, base_url=self.base_url)
        elif self.mfa_method in ('chip_tan_manual', 'chip_tan_qr'):
            mfa_auth = TANAuthentication(client=self.client, base_url=self.base_url, mfa_method=self.mfa_method)
        else:
            raise DKBRoboError(f'Login failed: unknown mfa method: {self.mfa_method}')

        mfa_completed = False
        if mfa_auth:
            mfa_completed = mfa_auth.finalize(challenge_id, challenge_dic, devicename)

        logger.debug('Authentication._mfa_finalize() ended with %s\n', mfa_completed)
        return mfa_completed

    def _mfa_get(self) -> Dict[str, str]:
        """ get mfa methods """
        logger.debug('Authentication._mfa_get()\n')
        mfa_dic = {}

        # check for access_token and get mfa_methods
        if 'access_token' in self.token_dic and 'mfa_id' in self.token_dic:
            response = self.client.get(self.base_url + f'/mfa/mfa/{self.token_dic["mfa_id"]}/methods?filter%5BmethodType%5D={self.mfa_method}')
            if response.status_code == 200:
                mfa_dic = response.json()
            else:
                raise DKBRoboError(f'Login failed: getting mfa_methods failed. RC: {response.status_code}')
        else:
            raise DKBRoboError('Login failed: no 1fa access token.')

        logger.debug('Authentication._mfa_get() ended\n')
        return mfa_dic

    def _mfa_process(self, device_num: int, device_list: List[int], _tmp_device_num: str, deviceselection_completed: bool) -> Tuple[int, bool]:
        logger.debug('Authentication._mfa_process(%s)', _tmp_device_num)
        try:
            # we are referring to an index in a list thus we need to lower the user input by 1
            if int(_tmp_device_num) - 1 in device_list:
                deviceselection_completed = True
                device_num = int(_tmp_device_num) - 1
            else:
                print('\nWrong input!')
        except Exception:
            print('\nInvalid input!')

        logger.debug('Authentication._mfa_process()\n ended')
        return device_num, deviceselection_completed

    def _mfa_select(self, mfa_dic: Dict[str, str]) -> int:
        """ pick mfa_device from dictionary """
        logger.debug('Authentication._mfa_select()')
        device_num = 0

        # adjust self.mfa_device if the user input is too high
        if 'data' in mfa_dic and len(mfa_dic['data']) < self.mfa_device:
            logger.warning('User submitted mfa_device number is invalid. Ingoring...')
            self.mfa_device = 0

        if self.mfa_device > 0:
            # we need to lower by one as we refer to an index in a list
            logger.debug('api.Wrapper._mfa_select(): using user submitted mfa_device number: %s', self.mfa_device)
            device_num = self.mfa_device - 1

        elif 'data' in mfa_dic and len(mfa_dic['data']) > 1:
            device_list = []
            deviceselection_completed = False
            while not deviceselection_completed:
                print('\nPick an authentication device from the below list:')
                # we have multiple devices to select
                for idx, device_dic in enumerate(mfa_dic['data']):
                    device_list.append(idx)
                    if 'attributes' in device_dic and 'deviceName' in device_dic['attributes']:
                        # we should start counting with 1 for the user
                        print(f"[{idx + 1}] - {device_dic['attributes']['deviceName']}")
                _tmp_device_num = input(':')

                device_num, deviceselection_completed = self._mfa_process(device_num, device_list, _tmp_device_num, deviceselection_completed)

        logger.debug('Authentication._mfa_select() ended with: %s', device_num)
        return device_num

    def _mfa_sort(self, mfa_dic):
        """ sort mfa methods """
        logger.debug('Authentication._mfa_sort()')

        mfa_list = mfa_dic['data']
        if self.mfa_method == 'seal_one':
            mfa_list.sort(key=lambda x: (-x['attributes']['preferredDevice'], x['attributes']['enrolledAt']))

        logger.debug('Authentication._mfa_sort() ended with: %s elements', len(mfa_list))
        return {'data': mfa_list}

    def _session_new(self):
        """ new request session for the api calls """
        logger.debug('Authentication._session_new()\n')

        headers = {
            'Accept-Language': 'en-US,en;q=0.5',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
            'Cache-Control': 'no-cache',
            'Connection': 'keep-alive',
            'DNT': '1',
            'Pragma': 'no-cache',
            'Sec-Fetch-Dest': 'document',
            'Sec-Fetch-Mode': 'navigate',
            'Sec-Fetch-Site': 'none',
            'te': 'trailers',
            'priority': 'u=0',
            'sec-gpc': '1',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0'}

        client = requests.session()
        client.headers = headers
        if self.proxies:
            client.proxies = self.proxies
            client.verify = False

        # get cookies
        client.get(self.base_url + '/login')

        # add csrf token
        if '__Host-xsrf' in client.cookies:
            headers['x-xsrf-token'] = client.cookies['__Host-xsrf']
            client.headers = headers

        logger.debug('Authentication._session_new() ended\n')
        return client

    def _sso_redirect(self):
        """  redirect to access legacy page """
        logger.debug('Authentication._sso_redirect()\n')

        data_dic = {'data': {'cookieDomain': '.dkb.de'}}
        self.client.headers['Content-Type'] = 'application/json'
        self.client.headers['Sec-Fetch-Dest'] = 'empty'
        self.client.headers['Sec-Fetch-Mode'] = 'cors'
        self.client.headers['Sec-Fetch-Site'] = 'same-origin'

        response = self.client.post(self.base_url + '/sso-redirect', data=json.dumps(data_dic))

        if response.status_code != 200 or response.text != 'OK':
            logger.error('SSO redirect failed. RC: %s text: %s', response.status_code, response.text)
        clientcookies = self.client.cookies

        legacywrappper = Legacywrapper()
        # pylint: disable=w0212
        self.dkb_br = legacywrappper._new_instance(clientcookies)
        logger.debug('Authentication._sso_redirect() ended.\n')

    def _token_get(self):
        """ get access token """
        logger.debug('Authentication._token_get()\n')

        # login via API
        data_dic = {'grant_type': 'banking_user_sca', 'username': self.dkb_user, 'password': self.dkb_password, 'sca_type': 'web-login'}
        response = self.client.post(self.base_url + '/token', data=data_dic)
        if response.status_code == 200:
            self.token_dic = response.json()
        else:
            raise DKBRoboError(f'Login failed: 1st factor authentication failed. RC: {response.status_code}')
        logger.debug('Authentication._token_get() ended\n')

    def _token_update(self):
        """ update token information with 2fa iformation """
        logger.debug('Authentication._token_update()\n')

        data_dic = {'grant_type': 'banking_user_mfa', 'mfa_id': self.token_dic['mfa_id'], 'access_token': self.token_dic['access_token']}
        response = self.client.post(self.base_url + '/token', data=data_dic)
        if response.status_code == 200:
            self.token_dic = response.json()
        else:
            raise DKBRoboError(f'Login failed: token update failed. RC: {response.status_code}')

    def login(self) -> Tuple[Dict, None]:
        """ login into DKB banking area and perform an sso redirect """
        logger.debug('Authentication.login()\n')

        mfa_dic = {}

        # create new session
        self.client = self._session_new()

        # get token for 1fa
        self._token_get()

        # get mfa methods
        mfa_dic = self._mfa_get()

        if mfa_dic:
            # sort mfa methods
            mfa_dic = self._mfa_sort(mfa_dic)

        # pick mfa device from list
        device_number = self._mfa_select(mfa_dic)

        # we need a challege-id for polling so lets try to get it
        mfa_challenge_dic = None
        if 'mfa_id' in self.token_dic and 'data' in mfa_dic:
            mfa_challenge_dic, device_name = self._mfa_challenge(mfa_dic, device_number)
        else:
            raise DKBRoboError('Login failed: no 1fa access token.')

        # lets complete 2fa
        mfa_completed = False
        if mfa_challenge_dic:
            mfa_completed = self._mfa_finalize(mfa_challenge_dic, device_name)
        else:
            raise DKBRoboError('Login failed: No challenge id.')

        # update token dictionary
        if mfa_completed and 'access_token' in self.token_dic:
            self._token_update()
        else:
            raise DKBRoboError('Login failed: mfa did not complete')

        if 'token_factor_type' not in self.token_dic:
            raise DKBRoboError('Login failed: token_factor_type is missing')

        if 'token_factor_type' in self.token_dic and self.token_dic['token_factor_type'] != '2fa':

            raise DKBRoboError('Login failed: 2nd factor authentication did not complete')

        # get account overview
        overview = Overview(client=self.client, unfiltered=self.unfiltered)
        self.account_dic = overview.get()

        # redirect to legacy page
        self._sso_redirect()
        logger.debug('Authentication.login() ended\n')
        return self.account_dic, None

    def logout(self):
        """ logout function """
        logger.debug('Authentication.logout()\n')


class APPAuthentication:
    """ APPAuthentication class """

    def __init__(self, client: requests.Session, base_url: str = BASE_URL):
        self.client = client
        self.base_url = base_url

    def _check(self, polling_dic: Dict[str, str], cnt: 1) -> bool:
        logger.debug('APPAuthentication._check()\n')

        if 'data' in polling_dic and 'attributes' in polling_dic['data'] and 'verificationStatus' in polling_dic['data']['attributes']:

            logger.debug('APPAuthentication._check: cnt %s got %s flag', cnt, polling_dic['data']['attributes']['verificationStatus'])

            mfa_completed = False
            if (polling_dic['data']['attributes']['verificationStatus']) == 'processed':
                mfa_completed = True
            elif (polling_dic['data']['attributes']['verificationStatus']) == 'processing':
                logger.info('Status: %s. Waiting for confirmation', polling_dic['data']['attributes']['verificationStatus'])
            elif (polling_dic['data']['attributes']['verificationStatus']) == 'canceled':
                raise DKBRoboError('2fa chanceled by user')
            else:
                logger.info('Unknown processing status: %s', polling_dic['data']['attributes']['verificationStatus'])
        else:
            raise DKBRoboError('Login failed: processing status format is other than expected')

        logger.debug('APPAuthentication._check() ended with: %s\n', mfa_completed)
        return mfa_completed

    def _print(self, devicename: str):
        """ 2fa confirmation message """
        logger.debug('api.Wrapper._print()\n')
        if devicename:
            print(f'check your banking app on "{devicename}" and confirm login...')
        else:
            print('check your banking app and confirm login...')

    def finalize(self, challenge_id: str, _challenge_dic: Dict[str, str], devicename: str) -> bool:
        """ wait for confirmation for the 2nd factor """
        logger.debug('APPAuthentication.finalize()\n')

        self._print(devicename)

        cnt = 0
        mfa_completed = False
        # we give us 50 seconds to press a button on the phone
        while cnt <= 10:
            response = self.client.get(self.base_url + f"/mfa/mfa/challenges/{challenge_id}")
            cnt += 1
            if response.status_code == 200:
                polling_dic = response.json()
                if 'data' in polling_dic and 'attributes' in polling_dic['data'] and 'verificationStatus' in polling_dic['data']['attributes']:
                    # check processing status
                    mfa_completed = self._check(polling_dic, cnt)
                    if mfa_completed:
                        break
                else:
                    logger.error('error parsing polling response: %s', polling_dic)
            else:
                logger.error('Polling request failed. RC: %s', response.status_code)
            time.sleep(5)

        logger.debug('APPAuthentication.finalize(): %s\n', mfa_completed)
        return mfa_completed


class TANAuthentication:
    """ TANAuthentication class """

    def __init__(self, client: requests.Session, base_url: str = BASE_URL, mfa_method: str = 'chip_tan_manual'):
        self.client = client
        self.base_url = base_url
        self.mfa_method = mfa_method

    def _image(self, qr_data: str) -> None:
        """ show qr code """
        logger.debug('TANAuthentication._image()\n')

        # pylint: disable=c0415
        from PIL import Image
        qr_data = qr_data.replace('data:image/png;base64,', '')
        qr_data = qr_data.replace(' ', '+')
        qr_data = base64.b64decode(qr_data)
        data = io.BytesIO()
        data.write(qr_data)
        img = Image.open(data)
        img.show()

        logger.debug('TANAuthentication._image() ended\n')

    def _print(self, challenge_dic: Dict[str, str]) -> str:
        """ print instructions for chip tan """
        logger.debug('TANAuthentication._print()\n')

        if 'data' in challenge_dic and 'attributes' in challenge_dic['data'] and 'chipTan' in challenge_dic['data']['attributes'] and 'qrData' in challenge_dic['data']['attributes']['chipTan']:
            my_thread = threading.Thread(target=self._image, args=(challenge_dic['data']['attributes']['chipTan']['qrData'], ))
            my_thread.daemon = True
            my_thread.start()
            my_thread.join(timeout=0.5)

        tan = None
        if 'data' in challenge_dic and 'attributes' in challenge_dic['data'] and 'chipTan' in challenge_dic['data']['attributes']:
            if 'headline' in challenge_dic['data']['attributes']['chipTan']:
                print(f"{challenge_dic['data']['attributes']['chipTan']['headline']}\n")
            if 'instructions' in challenge_dic['data']['attributes']['chipTan']:
                for idx, instruction in enumerate(challenge_dic['data']['attributes']['chipTan']['instructions'], start=1):
                    print(f'{idx}. {instruction}\n')

            tan = input("TAN: ")

        logger.debug('TANAuthentication._print() ended\n')
        return tan

    def finalize(self, challenge_id: str, challenge_dic: Dict[str, str], _device_name: str = None) -> bool:
        """ complete 2fa with chip tan manual """
        logger.debug('TANAuthentication.finalize()\n')

        tan = self._print(challenge_dic)

        data_dic = {"data": {"attributes": {"challengeResponse": tan, "methodType": self.mfa_method}, "type": "mfa-challenge"}}
        self.client.headers['Content-Type'] = JSON_CONTENT_TYPE
        self.client.headers["Accept"] = "application/vnd.api+json"

        response = self.client.post(self.base_url + f"/mfa/mfa/challenges/{challenge_id}", data=json.dumps(data_dic))
        mfa_completed = False
        if response.status_code <= 300:
            result_dic = response.json()
            if 'data' in result_dic and 'attributes' in result_dic['data'] and 'verificationStatus' in result_dic['data']['attributes'] and result_dic['data']['attributes']['verificationStatus'] == 'authorized':
                mfa_completed = True
        else:
            raise DKBRoboError(f'Login failed: 2fa failed. RC: {response.status_code} text: {response.text}')

        self.client.headers.pop('Content-Type')
        self.client.headers.pop('Accept')

        logger.debug('TANAuthentication.finalize() ended with %s\n', mfa_completed)
        return mfa_completed



""" Module for handling dkb transactions """
# pylint: disable=c0415, r0913, c0103
from typing import Dict, List, Tuple, Optional, Union
from dataclasses import dataclass, asdict
import logging
import requests
from dkb_robo.utilities import Amount, Person, get_dateformat, filter_unexpected_fields, ulal, BASE_URL

LEGACY_DATE_FORMAT, API_DATE_FORMAT = get_dateformat()
logger = logging.getLogger(__name__)


class ProductGroup:
    """ ProductGroup class"""

    def _uid2names(self, data_ele: Dict[str, str]) -> Dict[str, str]:
        """ create a dictionary containing id to product-name mapping """
        logger.debug('ProductGroup._uid2names()\n')

        product_settings_dic = {}
        portfolio_dic = data_ele.get('attributes', {}).get('productSettings', {})
        for product_data in portfolio_dic.values():
            if isinstance(product_data, dict):
                for uid, product_value in product_data.items():
                    if 'name' in product_value:
                        product_settings_dic[uid] = product_value['name']
            else:
                logger.warning('uid2name mapping failed. product data are not in dictionary format')

        logger.debug('ProductGroup._uid2names() ended\n')
        return product_settings_dic

    def _group(self, data_ele: Dict[str, str]) -> List[str]:
        """ create a list of products per group """
        logger.debug('ProductGroup._group()\n')

        product_group_list = []
        portfolio_dic = data_ele.get('attributes', {}).get('productGroups', {})
        for product_group in sorted(portfolio_dic.values(), key=lambda x: x['index']):
            id_dic = {}
            for _id_dic in product_group['products'].values():
                for uid in _id_dic:
                    id_dic[_id_dic[uid]['index']] = uid
            product_group_list.append({'name': product_group['name'], 'product_list': id_dic})

        logger.debug('ProductGroup._group()\n')
        return product_group_list

    def map(self, data_ele: Dict[str, str]) -> Tuple[Dict[str, str], Dict[str, str], int]:
        """ fetch data """
        logger.debug('ProductGroup.map()\n')

        # crate uid-name mapping and items per product group needed to sort the productgroup
        return self._uid2names(data_ele), self._group(data_ele)


class Overview:
    """ Overview class """

    def __init__(self, client: requests.Session, unfiltered: bool = False, base_url: str = BASE_URL):
        self.client = client
        self.base_url = base_url
        self.unfiltered = unfiltered

    def _add(self, data_dic: Dict[str, str], product_group: Dict[str, str], dic_id: str, product_display_dic: Dict[str, str]) -> Dict[str, str]:
        """ add product to account_dic """
        logger.debug('Overview._add()\n')

        # add product data to account_dic
        acc_dic = data_dic[product_group['product_list'][dic_id]]
        # add productgroup name to account_dic
        if self.unfiltered:
            acc_dic.productGroup = product_group['name']
        else:
            acc_dic['productgroup'] = product_group['name']

        if product_group['product_list'][dic_id] in product_display_dic:
            logger.debug('Overview._sort(): found displayname "%s" for product %s', product_display_dic[product_group['product_list'][dic_id]], product_group['product_list'][dic_id])
            # overwrite product name with display name
            if self.unfiltered:
                acc_dic.displayName = product_display_dic[product_group['product_list'][dic_id]]
            else:
                acc_dic['name'] = product_display_dic[product_group['product_list'][dic_id]]

        logger.debug('Overview._add() ended\n')
        return acc_dic

    def _add_remaining(self, data_dic: Dict[str, str], account_dic: Dict[str, str], account_cnt: int) -> Dict[str, str]:
        """ add remaining products """
        logger.debug('Overview._add_remaining()\n')

        for product_data in data_dic.values():
            account_dic[account_cnt] = product_data
            if self.unfiltered:
                account_dic[account_cnt].productGroup = None
            else:
                account_dic[account_cnt]['productgroup'] = None
            account_cnt += 1

        logger.debug('Overview._add_remaining() ended\n')
        return account_dic

    def _fetch(self, url_path) -> Dict[str, str]:
        """ fetch data via API """
        logger.debug('Overview._fetch()\n')

        response = self.client.get(self.base_url + url_path)
        if response.status_code == 200:
            response_dic = response.json()
        else:
            logger.error('fetch %s: RC is not 200 but %s', url_path, response.status_code)
            response_dic = {}

        logger.debug('Overview._fetch() ended\n')
        return response_dic

    def _sort(self, portfolio_dic: Dict[str, str]) -> Dict[str, str]:
        """ format and sort data """
        logger.debug('Overview._sort()\n')

        account_dic = {}
        account_cnt = 0

        data_dic = self._itemize(portfolio_dic)

        display_settings_dic = portfolio_dic.get('product_display', {}).get('data', {})
        productgroup = ProductGroup()
        for portfolio in display_settings_dic:
            # get id/name mapping and productlist per group
            product_display_dic, product_group_list = productgroup.map(portfolio)
            for product_group in product_group_list:
                # dic_id is a uid of the product
                for dic_id in sorted(product_group['product_list']):
                    if product_group['product_list'][dic_id] in data_dic:
                        logger.debug('Overview._sort(): assign productgroup "%s" to product %s', product_group['name'], product_group['product_list'][dic_id])
                        # add to dictionary
                        account_dic[account_cnt] = self._add(data_dic, product_group, dic_id, product_display_dic)
                        del data_dic[product_group['product_list'][dic_id]]
                        account_cnt += 1

        # add products without productgroup
        account_dic = self._add_remaining(data_dic, account_dic, account_cnt)

        logger.debug('Overview._sort() ended\n')
        return account_dic

    def _itemize(self, portfolio_dic: Dict[str, str]) -> Dict[str, str]:
        """ raw data """
        logger.debug('Overview._itemize()\n')

        product_dic = {}

        product_group_dic = {
            'accounts': 'AccountItem',
            'cards': 'CardItem',
            'depots': 'DepotItem'
        }
        for product_group in sorted(product_group_dic.keys()):
            if product_group in portfolio_dic and 'data' in portfolio_dic[product_group]:
                for item in portfolio_dic[product_group]['data']:
                    item['attributes']['id'] = item.get('id', None)
                    item['attributes']['type'] = item.get('type', None)
                    product = globals()[product_group_dic[product_group]](**item['attributes'])

                    if self.unfiltered:
                        product_dic[item['id']] = product
                    else:
                        product_dic[item['id']] = product.format()

        logger.debug('Overview._itemize() ended\n')
        return product_dic

    def get(self):
        """ Get overview """
        logger.debug('Overview.get()')

        # we calm the IDS system of DKB with two calls without sense
        # self._fetch('/terms-consent/consent-requests??filter%5Bportfolio%5D=DKB')
        product_display_dic = self._fetch('/config/users/me/product-display-settings')
        if product_display_dic:
            portfolio_dic = {
                'product_display': product_display_dic,
                'accounts': self._fetch('/accounts/accounts'),
                'cards': self._fetch('/credit-card/cards?filter%5Btype%5D=creditCard&filter%5Bportfolio%5D=dkb&filter%5Btype%5D=debitCard'),
                'depots': self._fetch('/broker/brokerage-accounts'),
                'loans': self._fetch('/loans/loans')
            }
        else:
            portfolio_dic = {}

        logger.debug('Overview.get() ended\n')
        return self._sort(portfolio_dic)


@filter_unexpected_fields
@dataclass
class AccountItem:
    """ Account dataclass """
    availableBalance: Optional[Union[Dict, str]] = None
    balance: Optional[Union[Dict, str]] = None
    currencyCode: Optional[str] = None
    displayName: Optional[str] = None
    holderName: Optional[str] = None
    id: Optional[str] = None
    iban: Optional[str] = None
    interestRate: Optional[Union[Dict, str]] = None
    interests: Optional[List] = None
    lastAccountStatementDate: Optional[str] = None
    nearTimeBalance: Optional[Union[Dict, str]] = None
    openingDate: Optional[str] = None
    overdraftLimit: Optional[Union[Dict, str]] = None
    permissions: Optional[List] = None
    product: Optional[Dict] = None
    productGroup: Optional[str] = None
    state: Optional[str] = None
    type: Optional[str] = None
    unauthorizedOverdraftInterestRate: Optional[Union[Dict, str]] = None
    updatedAt: Optional[str] = None
    transactions: Optional[str] = None

    def __post_init__(self):
        self.availableBalance = ulal(Amount, self.availableBalance)
        self.balance = ulal(Amount, self.balance)
        self.interests = [self.InterestsItem(**interest) for interest in self.interests]
        self.nearTimeBalance = ulal(Amount, self.nearTimeBalance)
        self.product = ulal(self.Product, self.product)
        self.transactions = BASE_URL + f"/accounts/accounts/{self.id}/transactions"
        try:
            self.overdraftLimit = float(self.overdraftLimit)
        except Exception:
            self.overdraftLimit = None

    @filter_unexpected_fields
    @dataclass
    class InterestsItem:
        """ interests class """
        details: Optional[List] = None
        method: Optional[str] = None
        type: Optional[str] = None

        def __post_init__(self):
            self.details = [self.DetailsItem(**detail) for detail in self.details]

        @filter_unexpected_fields
        @dataclass
        class DetailsItem:
            """ details class """
            condition: Optional[Union[Dict, str]] = None
            interestRate: Optional[float] = None

            def __post_init__(self):
                self.condition = ulal(self.Condition, self.condition)

            @filter_unexpected_fields
            @dataclass
            class Condition:
                """ condition class """
                currency: Optional[str] = None
                maximumAmount: Optional[float] = None
                minimumAmount: Optional[float] = None

                def __post_init__(self):
                    if self.minimumAmount:
                        try:
                            self.minimumAmount = float(self.minimumAmount)
                        except Exception:
                            self.minimumAmount = None
                    if self.maximumAmount:
                        try:
                            self.maximumAmount = float(self.maximumAmount)
                        except Exception:
                            self.maximumAmount = None

    @filter_unexpected_fields
    @dataclass
    class Product:
        """ Product class """
        id: Optional[str] = None
        type: Optional[str] = None
        displayName: Optional[str] = None

    def format(self) -> Dict[str, str]:
        """ format account """
        logger.debug('Account.format()\n')

        output_dic = {
            # for backward compatibility
            'account': self.iban,
            'amount': self.balance.value,
            'currencyCode': self.currencyCode,
            'date': self.updatedAt,
            'holderName': self.holderName,
            'iban': self.iban,
            'id': self.id,
            'limit': self.overdraftLimit,
            'name': self.product.displayName,
            'transactions': self.transactions,
            'type': self.type,
        }

        logger.debug('Account.format() ended\n')
        return output_dic


@filter_unexpected_fields
@dataclass
class CardItem:
    """ Card class """
    activationDate: Optional[str] = None
    authorizedAmount: Optional[Union[Dict, str]] = None
    availableLimit: Optional[Union[Dict, str]] = None
    balance: Optional[Union[Dict, str]] = None
    billingDetails: Optional[Union[Dict, str]] = None
    blockedSince: Optional[str] = None
    creationDate: Optional[str] = None
    engravedLine1: Optional[str] = None
    engravedLine2: Optional[str] = None
    expiryDate: Optional[str] = None
    failedPinAttempts: Optional[int] = None
    followUpCardId: Optional[str] = None
    holder: Optional[Union[Dict, str]] = None
    id: Optional[str] = None
    limit: Optional[Union[Dict, str]] = None
    maskedPan: Optional[str] = None
    network: Optional[str] = None
    owner: Optional[Union[Dict, str]] = None
    product: Optional[Union[Dict, str]] = None
    referenceAccount: Optional[str] = None
    state: Optional[str] = None
    status: Optional[Union[Dict, str]] = None
    type: Optional[str] = None
    transactions: Optional[str] = None

    def __post_init__(self):
        self.balance = ulal(Amount, self.balance)
        self.owner = ulal(Person, self.owner)
        self.availableLimit = ulal(Amount, self.availableLimit)
        self.authorizedAmount = ulal(Amount, self.authorizedAmount)
        self.referenceAccount = ulal(self.Account, self.referenceAccount)
        self.billingDetails = ulal(self.BillingDetails, self.billingDetails)
        self.limit = self.Limit(**self.limit)
        self.product = ulal(self.Product, self.product)
        self.status = ulal(self.Status, self.status)
        self.holder = ulal(self.Holder, self.holder)
        if self.type:
            self.transactions = BASE_URL + f"/card-transactions/creditcard-transactions?cardId={self.id}"

    @filter_unexpected_fields
    @dataclass
    class Account:
        """ Account class """
        iban: Optional[str] = None
        bic: Optional[str] = None

    @filter_unexpected_fields
    @dataclass
    class BillingDetails:
        """ BillingDetails class """
        days: Optional[List] = None
        calendarType: Optional[str] = None
        cycle: Optional[str] = None

    @filter_unexpected_fields
    @dataclass
    class Holder:
        """ Holder class """
        person: Optional[Dict] = None

        def __post_init__(self):
            self.person = ulal(Person, self.person)

    @filter_unexpected_fields
    @dataclass
    class Limit:
        """ Limit class """
        value: Optional[float] = None
        currencyCode: Optional[str] = None
        identifier: Optional[str] = None
        categories: Optional[List] = None

        def __post_init__(self):
            if self.value:
                try:
                    self.value = float(self.value)
                except Exception:
                    self.value = None
            if self.categories:
                self.categories = [self.CategoryItem(**category) for category in self.categories]

        @filter_unexpected_fields
        @dataclass
        class CategoryItem:
            """ Category class """
            amount: Optional[float] = None
            name: Optional[str] = None

            def __post_init__(self):
                self.amount = ulal(Amount, self.amount)

    @filter_unexpected_fields
    @dataclass
    class Product:
        """ Product class """
        superProductId: Optional[str] = None
        displayName: Optional[str] = None
        institute: Optional[str] = None
        productType: Optional[str] = None
        ownerType: Optional[str] = None
        id: Optional[str] = None
        type: Optional[str] = None

    @filter_unexpected_fields
    @dataclass
    class Status:
        """ Status class """
        category: Optional[str] = None
        since: Optional[str] = None
        reason: Optional[str] = None
        final: Optional[bool] = None
        limitationsFor: Optional[List] = None

    def format(self) -> Dict[str, str]:
        """ format card """
        logger.debug('Card.format()\n')

        output_dic = {
            'account': self.maskedPan,
            'expirydate': self.expiryDate,
            'holdername:': self.holder.person.firstName + ' ' + self.holder.person.lastName,
            'id': self.id,
            'name': self.product.displayName,
            'limit': self.limit.value,
            'maskedpan': self.maskedPan,
            'status': asdict(self.status),
            'type': self.type.lower(),

        }
        if self.type == 'creditCard':
            output_dic['transactions'] = self.transactions
            # dkb does some weird stuff with the balance. we need to flip it
            output_dic['amount'] = self.balance.value * -1
            output_dic['currencycode'] = self.balance.currencyCode
            output_dic['date'] = self.balance.date
        else:
            output_dic['transactions'] = None

        logger.debug('Card.format() ended\n')
        return output_dic


@filter_unexpected_fields
@dataclass
class DepotItem:
    """ Depot class """

    brokerageAccountPerformance: Optional[Union[Dict, str]] = None
    depositAccountId: Optional[str] = None
    holder: Optional[Union[Dict, str]] = None
    holderName: Optional[str] = None
    id: Optional[str] = None
    referenceAccounts: Optional[List] = None
    riskClasses: Optional[List] = None
    tradingEnabled: Optional[bool] = None
    type: Optional[str] = None
    transactions: Optional[str] = None

    def __post_init__(self):
        self.brokerageAccountPerformance = ulal(self.BrokerageAccountPerformance, self.brokerageAccountPerformance)
        self.holder = ulal(Person, self.holder)
        self.referenceAccounts = [self.ReferenceAccountItem(**reference_account) for reference_account in self.referenceAccounts]
        self.transactions = BASE_URL + f"/broker/brokerage-accounts/{self.id}/positions?include=instrument%2Cquote"

    @filter_unexpected_fields
    @dataclass
    class BrokerageAccountPerformance:
        """ BrokerageAccountPerformance class """
        currentValue: Optional[Union[Dict, str]] = None
        averagePrice: Optional[Union[Dict, str]] = None
        overallAbsolute: Optional[Union[Dict, str]] = None
        overallRelative: Optional[str] = None
        isOutdated: bool = False

        def __post_init__(self):
            self.currentValue = ulal(Amount, self.currentValue)
            self.averagePrice = ulal(Amount, self.averagePrice)
            self.overallAbsolute = ulal(Amount, self.overallAbsolute)

    @filter_unexpected_fields
    @dataclass
    class ReferenceAccountItem:
        """ ReferenceAccount class """
        internalReferenceAccounts: bool = False
        accountType: Optional[str] = None
        accountNumber: Optional[str] = None
        bankCode: Optional[str] = None
        holderName: Optional[str] = None

    def format(self) -> Dict[str, str]:
        """ format depot """
        logger.debug('Depot.format()\n')

        output_dic = {
            'account': self.depositAccountId,
            'amount': self.brokerageAccountPerformance.currentValue.value,
            'currencyCode': self.brokerageAccountPerformance.currentValue.currencyCode,
            'holderName': self.holderName,
            'id': self.id,
            'name': self.holderName,
            'type': 'depot',
            'transactions': self.transactions,
        }

        logger.debug('Depot.format() ended\n')
        return output_dic
